

Imports System.Diagnostics
Imports System.Runtime.InteropServices
Imports System.Threading
Imports LC.Multimedia

Public Partial Class InputDevice
	Inherits MidiDevice

    Private Sub HandleMessage(ByVal handle As Integer, ByVal msg As Integer, ByVal instance As Integer, ByVal param1 As Integer, ByVal param2 As Integer)
        If msg = MIM_OPEN Then
        ElseIf msg = MIM_CLOSE Then
        ElseIf msg = MIM_DATA Then
            delegateQueue.Post(AddressOf HandleShortMessage, param1)
        ElseIf msg = MIM_MOREDATA Then
            delegateQueue.Post(AddressOf HandleShortMessage, param1)
        ElseIf msg = MIM_LONGDATA Then
            delegateQueue.Post(AddressOf HandleSysExMessage, New IntPtr(param1))
        ElseIf msg = MIM_ERROR Then
            delegateQueue.Post(AddressOf HandleInvalidShortMessage, param1)
        ElseIf msg = MIM_LONGERROR Then
            delegateQueue.Post(AddressOf HandleInvalidSysExMessage, New IntPtr(param1))
        End If
    End Sub

	Private Sub HandleShortMessage(state As Object)
		Dim message As Integer = CInt(state)
		Dim status As Integer = ShortMessage.UnpackStatus(message)

		If status >= CInt(ChannelCommand.NoteOff) AndAlso status <= CInt(ChannelCommand.PitchWheel) + ChannelMessage.MidiChannelMaxValue Then
			cmBuilder.Message = message
			cmBuilder.Build()

			OnChannelMessageReceived(New ChannelMessageEventArgs(cmBuilder.Result))
		ElseIf status = CInt(SysCommonType.MidiTimeCode) OrElse status = CInt(SysCommonType.SongPositionPointer) OrElse status = CInt(SysCommonType.SongSelect) OrElse status = CInt(SysCommonType.TuneRequest) Then
			scBuilder.Message = message
			scBuilder.Build()

			OnSysCommonMessageReceived(New SysCommonMessageEventArgs(scBuilder.Result))
		Else
			Dim e As SysRealtimeMessageEventArgs = Nothing

			Select Case CType(status, SysRealtimeType)
                Case SysRealtimeType.MsgActiveSense
                    e = SysRealtimeMessageEventArgs.MsgActiveSense

                Case SysRealtimeType.MsgClock
                    e = SysRealtimeMessageEventArgs.MsgClock

                Case SysRealtimeType.MsgContinue
                    e = SysRealtimeMessageEventArgs.MsgContinue

                Case SysRealtimeType.MsgReset
                    e = SysRealtimeMessageEventArgs.MsgReset

                Case SysRealtimeType.MsgStart
                    e = SysRealtimeMessageEventArgs.MsgStart

                Case SysRealtimeType.MsgStop
                    e = SysRealtimeMessageEventArgs.MsgStop

                Case SysRealtimeType.MsgTick
                    e = SysRealtimeMessageEventArgs.MsgTick

            End Select

			OnSysRealtimeMessageReceived(e)
		End If
	End Sub

	Private Sub HandleSysExMessage(state As Object)
		SyncLock lockObject
			Dim headerPtr As IntPtr = CType(state, IntPtr)

			Dim header As MidiHeader = CType(Marshal.PtrToStructure(headerPtr, GetType(MidiHeader)), MidiHeader)

			If Not resetting Then
				For i As Integer = 0 To header.bytesRecorded - 1
					sysExData.Add(Marshal.ReadByte(header.data, i))
				Next

				If sysExData(0) = &Hf0 AndAlso sysExData(sysExData.Count - 1) = &Hf7 Then
					Dim message As New SysExMessage(sysExData.ToArray())

					sysExData.Clear()

					OnSysExMessageReceived(New SysExMessageEventArgs(message))
				End If

				Dim result As Integer = AddSysExBuffer()

				If result <> DeviceException.MMSYSERR_NOERROR Then
					Dim ex As Exception = New InputDeviceException(result)

					OnError(New ErrorEventArgs(ex))
				End If
			End If

			ReleaseBuffer(headerPtr)
		End SyncLock
	End Sub

	Private Sub HandleInvalidShortMessage(state As Object)
		OnInvalidShortMessageReceived(New InvalidShortMessageEventArgs(CInt(state)))
	End Sub

	Private Sub HandleInvalidSysExMessage(state As Object)
		SyncLock lockObject
			Dim headerPtr As IntPtr = CType(state, IntPtr)

			Dim header As MidiHeader = CType(Marshal.PtrToStructure(headerPtr, GetType(MidiHeader)), MidiHeader)

			If Not resetting Then
				Dim data As Byte() = New Byte(header.bytesRecorded - 1) {}

				Marshal.Copy(header.data, data, 0, data.Length)

				OnInvalidSysExMessageReceived(New InvalidSysExMessageEventArgs(data))

				Dim result As Integer = AddSysExBuffer()

				If result <> DeviceException.MMSYSERR_NOERROR Then
					Dim ex As Exception = New InputDeviceException(result)

					OnError(New ErrorEventArgs(ex))
				End If
			End If

			ReleaseBuffer(headerPtr)
		End SyncLock
	End Sub

	Private Sub ReleaseBuffer(headerPtr As IntPtr)
        Dim result As Integer = midiInUnprepareHeader(m_handle, headerPtr, SizeOfMidiHeader)

		If result <> DeviceException.MMSYSERR_NOERROR Then
			Dim ex As Exception = New InputDeviceException(result)

			OnError(New ErrorEventArgs(ex))
		End If

		headerBuilder.Destroy(headerPtr)

		bufferCount -= 1

		Debug.Assert(bufferCount >= 0)

		Monitor.Pulse(lockObject)
	End Sub

	Public Function AddSysExBuffer() As Integer
		Dim result As Integer

		' Initialize the MidiHeader builder.
        headerBuilder.BufferLength = m_sysExBufferSize
		headerBuilder.Build()

		' Get the pointer to the built MidiHeader.
		Dim headerPtr As IntPtr = headerBuilder.Result

		' Prepare the header to be used.
        result = midiInPrepareHeader(m_handle, headerPtr, SizeOfMidiHeader)

		' If the header was perpared successfully.
		If result = DeviceException.MMSYSERR_NOERROR Then
			bufferCount += 1

			' Add the buffer to the InputDevice.
            result = midiInAddBuffer(m_handle, headerPtr, SizeOfMidiHeader)

			' If the buffer could not be added.
			If result <> MidiDeviceException.MMSYSERR_NOERROR Then
				' Unprepare header - there's a chance that this will fail 
				' for whatever reason, but there's not a lot that can be
				' done at this point.
                midiInUnprepareHeader(m_handle, headerPtr, SizeOfMidiHeader)

				bufferCount -= 1

				' Destroy header.
				headerBuilder.Destroy()
			End If
		Else
			' Else the header could not be prepared.
			' Destroy header.
			headerBuilder.Destroy()
		End If

		Return result
    End Function

End Class
