

Imports System.Collections.Generic
Imports System.Runtime.InteropServices
Imports System.Threading
Imports LC.Timers

Public NotInheritable Class OutputStream
	Inherits OutputDeviceBase
	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamOpen(ByRef handle As Integer, ByRef deviceID As Integer, reserved As Integer, proc As OutputDevice.MidiOutProc, instance As Integer, flag As UInteger) As Integer
	End Function

	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamClose(handle As Integer) As Integer
	End Function

	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamOut(handle As Integer, headerPtr As IntPtr, sizeOfMidiHeader As Integer) As Integer
	End Function

	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamPause(handle As Integer) As Integer
	End Function

	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamPosition(handle As Integer, ByRef t As Time, sizeOfTime As Integer) As Integer
	End Function

    <DllImport("winmm.dll")> _
    Private Shared Function midiStreamProperty(ByVal handle As Integer, ByRef p As [Property], ByVal flags As Int32) As Integer
    End Function

	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamRestart(handle As Integer) As Integer
	End Function

	<DllImport("winmm.dll")> _
	Private Shared Function midiStreamStop(handle As Integer) As Integer
	End Function

	<StructLayout(LayoutKind.Sequential)> _
	Private Structure [Property]
		Public sizeOfProperty As Integer
		Public [property] As Integer
	End Structure

    Private Const MIDIPROP_SET As Int32 = &H80000000  ' TODO LIVIO Changed fron Uint32 ( ok ?? )
    Private Const MIDIPROP_GET As Int32 = &H40000000  ' TODO LIVIO Changed fron Uint32 ( ok ?? )
	Private Const MIDIPROP_TIMEDIV As UInteger = &H1
	Private Const MIDIPROP_TEMPO As UInteger = &H2

	Private Const MEVT_CALLBACK As Byte = &H40

	Private Const MEVT_SHORTMSG As Byte = &H0
	Private Const MEVT_TEMPO As Byte = &H1
	Private Const MEVT_NOP As Byte = &H2
	Private Const MEVT_LONGMSG As Byte = &H80
	Private Const MEVT_COMMENT As Byte = &H82
	Private Const MEVT_VERSION As Byte = &H84

	Private Const MOM_POSITIONCB As Integer = &H3ca

	Private Const SizeOfMidiEvent As Integer = 12

	Private Const EventTypeIndex As Integer = 11

	Private Const EventCodeOffset As Integer = 8

    Private m_midiOutProc As MidiOutProc

	Private offsetTicks As Integer = 0

	Private streamID As Byte() = New Byte(3) {}

	Private events As New List(Of Byte)()

	Private headerBuilder As New MidiHeaderBuilder()

	Public Event NoOpOccurred As EventHandler(Of NoOpEventArgs)

	Public Sub New(deviceID As Integer)
		MyBase.New(deviceID)
        m_midiOutProc = AddressOf HandleMessage

        Dim result As Integer = midiStreamOpen(m_handle, deviceID, 1, m_midiOutProc, 0, CALLBACK_FUNCTION)

		If result <> MidiDeviceException.MMSYSERR_NOERROR Then
			Throw New OutputDeviceException(result)
		End If
	End Sub

	Protected Overloads Overrides Sub Dispose(disposing As Boolean)
		If disposing Then
			SyncLock lockObject
				Reset()

                Dim result As Integer = midiStreamClose(m_handle)

				If result <> MidiDeviceException.MMSYSERR_NOERROR Then
					Throw New OutputDeviceException(result)
				End If
			End SyncLock
		Else
            midiOutReset(m_handle)
            midiStreamClose(m_handle)
		End If

		MyBase.Dispose(disposing)
	End Sub

	Public Overloads Overrides Sub Close()
        If Not IsDisposed Then
            Return
        End If
        Dispose(True)
    End Sub

	Public Sub StartPlaying()
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If
        SyncLock lockObject
            Dim result As Integer = midiStreamRestart(m_handle)
            If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                Throw New OutputDeviceException(result)
            End If
        End SyncLock
    End Sub

	Public Sub PausePlaying()
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If
        SyncLock lockObject
            Dim result As Integer = midiStreamPause(m_handle)

            If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                Throw New OutputDeviceException(result)
            End If
        End SyncLock
    End Sub

	Public Sub StopPlaying()
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If
        SyncLock lockObject
            Dim result As Integer = midiStreamStop(m_handle)

            If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                Throw New OutputDeviceException(result)
            End If
        End SyncLock
    End Sub

	Public Overloads Overrides Sub Reset()
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If
        offsetTicks = 0
        events.Clear()
        MyBase.Reset()
    End Sub

	Public Sub Write(e As MidiEvent)
		Select Case e.MidiMessage.MessageType
			Case MessageType.Channel, MessageType.SystemCommon, MessageType.SystemRealtime
				Write(e.DeltaTicks, DirectCast(e.MidiMessage, ShortMessage))

			Case MessageType.SystemExclusive
				Write(e.DeltaTicks, DirectCast(e.MidiMessage, SysExMessage))

			Case MessageType.Meta
				Write(e.DeltaTicks, DirectCast(e.MidiMessage, MetaMessage))

		End Select
	End Sub

	Private Sub Write(deltaTicks As Integer, message As ShortMessage)
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If

        ' Delta time.
        events.AddRange(BitConverter.GetBytes(deltaTicks + offsetTicks))

        ' Stream ID.
        events.AddRange(streamID)

        ' Event code.
        Dim eventCode As Byte() = message.GetBytes()
        eventCode(eventCode.Length - 1) = MEVT_SHORTMSG
        events.AddRange(eventCode)

        offsetTicks = 0
    End Sub

	Private Sub Write(deltaTicks As Integer, message As SysExMessage)
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If

        ' Delta time.
        events.AddRange(BitConverter.GetBytes(deltaTicks + offsetTicks))

        ' Stream ID.
        events.AddRange(streamID)

        ' Event code.
        Dim eventCode As Byte() = BitConverter.GetBytes(message.Length)
        eventCode(eventCode.Length - 1) = MEVT_LONGMSG
        events.AddRange(eventCode)

        Dim sysExData As Byte()

        If message.Length Mod 4 <> 0 Then
            sysExData = New Byte(message.Length + ((message.Length Mod 4) - 1)) {}
            message.GetBytes().CopyTo(sysExData, 0)
        Else
            sysExData = message.GetBytes()
        End If

        ' SysEx data.
        events.AddRange(sysExData)

        offsetTicks = 0
    End Sub

	Private Sub Write(deltaTicks As Integer, message As MetaMessage)
		If message.MetaType = MetaType.Tempo Then
			' Delta time.
			events.AddRange(BitConverter.GetBytes(deltaTicks + offsetTicks))

			' Stream ID.
			events.AddRange(streamID)

			Dim builder As New TempoChangeBuilder(message)

			Dim t As Byte() = BitConverter.GetBytes(builder.Tempo)

			t(t.Length - 1) = MEVT_SHORTMSG Or MEVT_TEMPO

			' Event code.
			events.AddRange(t)

			offsetTicks = 0
		Else
			offsetTicks += deltaTicks
		End If
	End Sub

	Public Sub WriteNoOp(deltaTicks As Integer, data As Integer)
		' Delta time.
		events.AddRange(BitConverter.GetBytes(deltaTicks + offsetTicks))

		' Stream ID.
		events.AddRange(streamID)

		' Event code.
		Dim eventCode As Byte() = BitConverter.GetBytes(data)
		eventCode(eventCode.Length - 1) = CByte((MEVT_NOP Or MEVT_CALLBACK))
		events.AddRange(eventCode)

		offsetTicks = 0
	End Sub

	Public Sub Flush()
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If
        SyncLock lockObject
            headerBuilder.InitializeBuffer(events)
            headerBuilder.Build()

            events.Clear()

            Dim result As Integer = midiOutPrepareHeader(m_handle, headerBuilder.Result, SizeOfMidiHeader)

            If result = MidiDeviceException.MMSYSERR_NOERROR Then
                bufferCount += 1
            Else
                headerBuilder.Destroy()

                Throw New OutputDeviceException(result)
            End If

            result = midiStreamOut(m_handle, headerBuilder.Result, SizeOfMidiHeader)

            If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                midiOutUnprepareHeader(m_handle, headerBuilder.Result, SizeOfMidiHeader)

                headerBuilder.Destroy()

                Throw New OutputDeviceException(result)
            End If
        End SyncLock
    End Sub

	Public Function GetTime(type As TimeType) As Time
        If IsDisposed Then
            Throw New ObjectDisposedException("OutputStream")
        End If

        Dim t As New Time()

        t.type = type

        SyncLock lockObject
            Dim result As Integer = midiStreamPosition(m_handle, t, Marshal.SizeOf(GetType(Time)))

            If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                Throw New OutputDeviceException(result)
            End If
        End SyncLock

        Return t
    End Function

	Private Sub OnNoOpOccurred(e As NoOpEventArgs)
        RaiseEvent NoOpOccurred(Me, e)
    End Sub

	Protected Overloads Overrides Sub HandleMessage(handle As Integer, msg As Integer, instance As Integer, param1 As Integer, param2 As Integer)
		If msg = MOM_POSITIONCB Then
			delegateQueue.Post(AddressOf HandleNoOp, New IntPtr(param1))
		Else
			MyBase.HandleMessage(handle, msg, instance, param1, param2)
		End If
	End Sub

	Private Sub HandleNoOp(state As Object)
		Dim headerPtr As IntPtr = CType(state, IntPtr)
		Dim header As MidiHeader = CType(Marshal.PtrToStructure(headerPtr, GetType(MidiHeader)), MidiHeader)

		Dim midiEvent As Byte() = New Byte(SizeOfMidiEvent - 1) {}

		For i As Integer = 0 To midiEvent.Length - 1
			midiEvent(i) = Marshal.ReadByte(header.data, header.offset + i)
		Next

		' If this is a NoOp event.
		If (midiEvent(EventTypeIndex) And MEVT_NOP) = MEVT_NOP Then
			' Clear the event type byte.
			midiEvent(EventTypeIndex) = 0

            OnNoOpOccurredEventArgs = New NoOpEventArgs(BitConverter.ToInt32(midiEvent, EventCodeOffset))

            context.Post(New SendOrPostCallback(AddressOf SendOrPostCallbackDelegate), Nothing)
        End If
    End Sub
    Private OnNoOpOccurredEventArgs As NoOpEventArgs
    Private Sub SendOrPostCallbackDelegate(ByVal s As Object)
        OnNoOpOccurred(OnNoOpOccurredEventArgs)
    End Sub


    Public Property Division() As Integer
        Get
            If IsDisposed Then
                Throw New ObjectDisposedException("OutputStream")
            End If

            Dim d As New [Property]()

            d.sizeOfProperty = Marshal.SizeOf(GetType([Property]))

            SyncLock lockObject
                Dim result As Integer = midiStreamProperty(m_handle, d, MIDIPROP_GET Or MIDIPROP_TIMEDIV)

                If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                    Throw New OutputDeviceException(result)
                End If
            End SyncLock

            Return d.[property]
        End Get

        Set(ByVal value As Integer)
            If IsDisposed Then
                Throw New ObjectDisposedException("OutputStream")
            ElseIf (value Mod PpqnClock.PpqnMinValue) <> 0 Then
                Throw New ArgumentException()
            End If

            Dim d As New [Property]()

            d.sizeOfProperty = Marshal.SizeOf(GetType([Property]))
            d.[property] = value

            SyncLock lockObject
                Dim result As Integer = midiStreamProperty(m_handle, d, MIDIPROP_SET Or MIDIPROP_TIMEDIV)

                If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                    Throw New OutputDeviceException(result)
                End If
            End SyncLock
        End Set
    End Property

    Public Property Tempo() As Integer
        Get
            If IsDisposed Then
                Throw New ObjectDisposedException("OutputStream")
            End If

            Dim t As New [Property]()
            t.sizeOfProperty = Marshal.SizeOf(GetType([Property]))

            SyncLock lockObject
                Dim result As Integer = midiStreamProperty(m_handle, t, MIDIPROP_GET Or MIDIPROP_TEMPO)

                If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                    Throw New OutputDeviceException(result)
                End If
            End SyncLock

            Return t.[property]
        End Get
        Set(ByVal value As Integer)
            If IsDisposed Then
                Throw New ObjectDisposedException("OutputStream")
            ElseIf value < 0 Then
                Throw New ArgumentOutOfRangeException("Tempo", value, "Tempo out of range.")
            End If

            Dim t As New [Property]()
            t.sizeOfProperty = Marshal.SizeOf(GetType([Property]))
            t.[property] = value

            SyncLock lockObject
                Dim result As Integer = midiStreamProperty(m_handle, t, MIDIPROP_SET Or MIDIPROP_TEMPO)

                If result <> MidiDeviceException.MMSYSERR_NOERROR Then
                    Throw New OutputDeviceException(result)
                End If
            End SyncLock
        End Set
    End Property

End Class
