

Imports System.ComponentModel
Imports System.Diagnostics



Public Enum SysCommonType
    MidiTimeCode = &HF1
    SongPositionPointer
    SongSelect
    TuneRequest = &HF6
End Enum


' ------------------------------------------------------------------ MIDI system common messages.
<ImmutableObject(True)> _
Public NotInheritable Class SysCommonMessage
    Inherits ShortMessage

    Public Sub New(ByVal type As SysCommonType)
        msg = type
        Debug.Assert(SysCommonType = type)
    End Sub

    Public Sub New(ByVal type As SysCommonType, ByVal data1__1 As Integer)
        msg = type
        msg = PackData1(msg, data1__1)
        Debug.Assert(SysCommonType = type)
        Debug.Assert(Data1 = data1__1)
    End Sub

    Public Sub New(ByVal type As SysCommonType, ByVal data1__1 As Integer, ByVal data2__2 As Integer)
        msg = type
        msg = PackData1(msg, data1__1)
        msg = PackData2(msg, data2__2)

        Debug.Assert(SysCommonType = type)
        Debug.Assert(Data1 = data1__1)
        Debug.Assert(Data2 = data2__2)
    End Sub

    Friend Sub New(ByVal message As Integer)
        Me.msg = message
    End Sub


    ' ------------------------------------------------------------------------------------------- "Methods"

    Public Overloads Overrides Function GetHashCode() As Integer
        Return msg
    End Function


    ' ----------------------------------------------------------- test if two SysCommonMessage instances are equal
    '
    Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
        If Not (TypeOf obj Is SysCommonMessage) Then
            Return False
        End If

        Dim message As SysCommonMessage = DirectCast(obj, SysCommonMessage)

        Return (Me.SysCommonType = message.SysCommonType AndAlso Me.Data1 = message.Data1 AndAlso Me.Data2 = message.Data2)
    End Function


    ' ---------------------------------------------------------------------------------------------- "Properties"


    ' ------------------------------------------------------------ gets the SysCommonType
    Public ReadOnly Property SysCommonType() As SysCommonType
        Get
            Return CType(UnpackStatus(msg), SysCommonType)
        End Get
    End Property

    ' ------------------------------------------------------------- gets the first data value
    Public ReadOnly Property Data1() As Integer
        Get
            Return UnpackData1(msg)
        End Get
    End Property

    ' ------------------------------------------------------------- gets the second data value
    Public ReadOnly Property Data2() As Integer
        Get
            Return UnpackData2(msg)
        End Get
    End Property

    ' -------------------------------------------------------------- gets the MessageType
    '
    Public Overloads Overrides ReadOnly Property MessageType() As MessageType
        Get
            Return MessageType.SystemCommon
        End Get
    End Property

End Class
