

Imports System.Collections

Public Enum SysExType
    Start = &HF0
    Continuation = &HF7
End Enum


' ---------------------------------------------------------------------------- MIDI system exclusive messages
Public NotInheritable Class SysExMessage
	Implements IMidiMessage
	Implements IEnumerable

    Public Const SysExChannelMaxValue As Integer = 127

    Private data As Byte()

    Public Sub New(ByVal data As Byte())
        If data.Length < 1 Then
            Throw New ArgumentException("System exclusive data is too short.", "data")
        ElseIf data(0) <> CByte(SysExType.Start) AndAlso data(0) <> CByte(SysExType.Continuation) Then
            Throw New ArgumentException("Unknown status value.", "data")
        End If
        Me.data = New Byte(data.Length - 1) {}
        data.CopyTo(Me.data, 0)
    End Sub


    ' ----------------------------------------------------------------------------------------- "Methods"

    Public Function GetBytes() As Byte() Implements IMidiMessage.GetBytes
        Dim clone As Byte() = New Byte(data.Length - 1) {}
        data.CopyTo(clone, 0)
        Return clone
    End Function

    Public Sub CopyTo(ByVal buffer As Byte(), ByVal index As Integer)
        data.CopyTo(buffer, index)
    End Sub

    Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
        If Not (TypeOf obj Is SysExMessage) Then
            Return False
        End If

        Dim message As SysExMessage = DirectCast(obj, SysExMessage)

        Dim _equals As Boolean = True

        If Me.Length <> message.Length Then
            _equals = False
        End If

        Dim i As Integer = 0
        While i < Me.Length AndAlso _equals
            If Me(i) <> message(i) Then
                _equals = False
            End If
            i += 1
        End While

        Return _equals
    End Function

    Public Overloads Overrides Function GetHashCode() As Integer
        Return data.GetHashCode()
    End Function


    ' -------------------------------------------------------------------------------------- "Properties"

    Default Public ReadOnly Property Item(ByVal index As Integer) As Byte
        Get
            If index < 0 OrElse index >= Length Then
                Throw New ArgumentOutOfRangeException("index", index, "Index into system exclusive message out of range.")
            End If
            Return data(index)
        End Get
    End Property

    Public ReadOnly Property Length() As Integer
        Get
            Return data.Length
        End Get
    End Property

    Public ReadOnly Property SysExType() As SysExType
        Get
            Return CType(data(0), SysExType)
        End Get
    End Property

	Public ReadOnly Property Status() As Integer Implements IMidiMessage.Status
		Get
            Return data(0)
		End Get
	End Property

	Public ReadOnly Property MessageType() As MessageType Implements IMidiMessage.MessageType
		Get
			Return MessageType.SystemExclusive
		End Get
	End Property

#Region "IEnumerable Members"

    Public Function GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        Return data.GetEnumerator()
    End Function

#End Region

End Class
