

Imports System.ComponentModel
Imports System.Diagnostics

Public Enum SysRealtimeType
    MsgClock = &HF8
    MsgTick
    MsgStart
    MsgContinue
    MsgStop
    MsgActiveSense = &HFE
    MsgReset
End Enum

' ---------------------------------------------------------------------------- system realtime messages
<ImmutableObject(True)> _
Public NotInheritable Class SysRealtimeMessage
    Inherits ShortMessage

    ' --------------------------------------------------------------
    '  SysRealtimeMessage INSTANCES
    ' --------------------------------------------------------------

    Public Shared ReadOnly StartMessage As New SysRealtimeMessage(SysRealtimeType.MsgStart)
    Public Shared ReadOnly ContinueMessage As New SysRealtimeMessage(SysRealtimeType.MsgContinue)
    Public Shared ReadOnly StopMessage As New SysRealtimeMessage(SysRealtimeType.MsgStop)
    Public Shared ReadOnly ClockMessage As New SysRealtimeMessage(SysRealtimeType.MsgClock)
    Public Shared ReadOnly TickMessage As New SysRealtimeMessage(SysRealtimeType.MsgTick)
    Public Shared ReadOnly ActiveSenseMessage As New SysRealtimeMessage(SysRealtimeType.MsgActiveSense)
    Public Shared ReadOnly ResetMessage As New SysRealtimeMessage(SysRealtimeType.MsgReset)

    ' ----------------- private construction so that a system realtime message cannot be constructed directly
    '
    Private Sub New(ByVal type As SysRealtimeType)
        msg = type
        Debug.Assert(SysRealtimeType = type)
    End Sub

    ' -------------------------------------------------------------------------------------------------- "Methods"

    Public Overloads Overrides Function GetHashCode() As Integer
        Return msg
    End Function

    Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
        If Not (TypeOf obj Is SysRealtimeMessage) Then
            Return False
        End If
        Dim message As SysRealtimeMessage = DirectCast(obj, SysRealtimeMessage)
        Return Me.msg = message.msg
    End Function


    ' ----------------------------------------------------------------------------------------------- "Properties"
    Public ReadOnly Property SysRealtimeType() As SysRealtimeType
        Get
            Return CType(msg, SysRealtimeType)
        End Get
    End Property

    Public Overloads Overrides ReadOnly Property MessageType() As MessageType
        Get
            Return MessageType.SystemRealtime
        End Get
    End Property

End Class
