

' -------------------------------------------------- functionality for raising 2 to a power by using lookup tables
Public NotInheritable Class PowerOfTwoTable

    Private Const TableSize As Integer = 4096
    Private Shared ReadOnly Table As Single() = New Single(TableSize - 1) {}


    Shared Sub New()
        ' ------------------------------------------ Calculate phase increment.
        Dim increment As Single = 1.0F / TableSize
        ' ------------------------------------------ Phase accumulator.
        Dim accumulator As Single = 0
        ' ------------------------------------------ Initialize lookup table.
        For i As Integer = 0 To TableSize - 1
            Table(i) = CSng(Math.Pow(2, accumulator))
            accumulator += increment
        Next
    End Sub


    ' ---------------------------------------------- Gets the result of 2 raised to the specified power.
    '
    Public Shared Function GetPower(ByVal exponent As Single) As Single
        Dim result As Single

        ' ----------------------------------------------------------- If exponent is positive.
        If exponent >= 0 Then
            ' ------------------------------------------------------- Get whole part of exponent.
            Dim whole As Integer = CInt(Math.Truncate(exponent))
            ' ------------------------------------------------------- Get fractional part of exponent.
            Dim fractional As Single = exponent - whole
            ' -------------------------------------------------------- Get index into lookup table. 
            '                                                           Some form of interpolation would help here,
            '                                                           but with a large enough lookup table,
            '                                                           the results seem acceptable.
            Dim index As Integer = CInt(Math.Truncate(TableSize * fractional))
            ' -------------------------------------------------------- Calculate final result. 
            result = Table(index) * (1 << whole)
        Else
            ' ------------------------------------------------------- Get whole part of exponent.
            Dim whole As Integer = CInt(Math.Truncate(-exponent))
            ' ------------------------------------------------------- Get fractional part of exponent.
            Dim fractional As Single = -exponent - whole
            ' ------------------------------------------------------- Get index into lookup table. 
            '                                                           Some form of interpolation would help here,
            '                                                           but with a large enough lookup table,
            '                                                           the results seem acceptable.
            Dim index As Integer = CInt(Math.Truncate(TableSize * fractional))

            ' ------------------------------------------------------- Calculate final result 
            '                                                          divide 1 by result since exponent is negative 
            result = 1.0F / (Table(index) * (1 << whole))
        End If

        Return result
    End Function

End Class
