

Imports System.Collections
Imports System.Collections.Generic
Imports System.Diagnostics

' ------------------------------------------------------------------------ a collection of IProgrammable objects
Public Class ProgrammableCollection
    Implements ICollection(Of IProgrammable)
    Implements IProgrammable

    ' -------------------------------------------------------------------- the collection of IProgrammable objects
    Private Programmables As New List(Of IProgrammable)()
    ' -------------------------------------------------------------------- the total number of parameters
    Private m_parameterCount As Integer = 0



    ' -------------------------------------------------------------------- map the index to a IProgrammable object
    Private Function MapToParameter(ByVal index As Integer, _
                                    ByRef Programmable As IProgrammable) As Integer

        Debug.Assert(index < ParameterCount)

        Dim i As Integer = index
        Dim en As IEnumerator(Of IProgrammable) = GetEnumerator()

        en.MoveNext()

        ' ---------------------------------------------------------------- Iterate through the collection until
        '                                                                  the index falls within the range 
        '                                                                  of one of the IProgrammable objects.
        While i >= en.Current.ParameterCount
            ' ------------------------------------------------------------ offset index
            i -= en.Current.ParameterCount
            en.MoveNext()
        End While
        Programmable = en.Current
        ' ---------------------------------------------------------------- return the offset index
        Return i
    End Function


    Public Sub Add(ByVal item As IProgrammable) Implements System.Collections.Generic.ICollection(Of IProgrammable).Add
        Programmables.Add(item)
        m_parameterCount += item.ParameterCount
    End Sub

    Public Sub Clear() Implements ICollection(Of IProgrammable).Clear
        Programmables.Clear()
        m_parameterCount = 0
    End Sub

    Public Function Contains(ByVal item As IProgrammable) As Boolean Implements System.Collections.Generic.ICollection(Of IProgrammable).Contains
        Return Programmables.Contains(item)
    End Function

    Public Sub CopyTo(ByVal array() As IProgrammable, ByVal arrayIndex As Integer) Implements System.Collections.Generic.ICollection(Of IProgrammable).CopyTo
        Programmables.CopyTo(array, arrayIndex)
    End Sub

    Public ReadOnly Property Count() As Integer Implements ICollection(Of IProgrammable).Count
        Get
            Return Programmables.Count
        End Get
    End Property

    Public ReadOnly Property IsReadOnly() As Boolean Implements ICollection(Of IProgrammable).IsReadOnly
        Get
            Return False
        End Get
    End Property

    Public Function Remove(ByVal item As IProgrammable) As Boolean Implements System.Collections.Generic.ICollection(Of IProgrammable).Remove
        Dim result As Boolean

        If Programmables.Remove(item) Then
            m_parameterCount -= item.ParameterCount

            result = True
        Else
            result = False
        End If

        Return result
    End Function


#Region "IEnumerable<IProgrammable> Members"

    Public Function GetEnumerator() As IEnumerator(Of IProgrammable) Implements IEnumerable(Of IProgrammable).GetEnumerator
        Return Programmables.GetEnumerator()
    End Function

#End Region

#Region "IEnumerable Members"

    Private Function IEnumerable_GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        Return Programmables.GetEnumerator()
    End Function

#End Region

#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim Programmable As IProgrammable = Nothing
        Dim i As Integer = MapToParameter(index, Programmable)

        Return Programmable.GetParameterName(i)
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim Programmable As IProgrammable = Nothing
        Dim i As Integer = MapToParameter(index, Programmable)

        Return Programmable.GetParameterLabel(i)
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim Programmable As IProgrammable = Nothing
        Dim i As Integer = MapToParameter(index, Programmable)

        Return Programmable.GetParameterDisplay(i)
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim Programmable As IProgrammable = Nothing
        Dim i As Integer = MapToParameter(index, Programmable)

        Return Programmable.GetParameterValue(i)
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If

        Dim Programmable As IProgrammable = Nothing
        Dim i As Integer = MapToParameter(index, Programmable)

        Programmable.SetParameterValue(i, value)
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return m_parameterCount
        End Get
    End Property

#End Region


End Class
