

Imports System.Diagnostics

Public Class StateVariableFilter
	Inherits MonoSynthComponent
	Implements IProgrammable

    Public Enum ParameterId
        CutoffFrequency
        Resonance
        FilterType
        KeyboardTracking
        FMLevel1
        FMPolarity1
        FMLevel2
        FMPolarity2
        Drive
    End Enum

    Public Enum FilterType
        Lowpass
        Highpass
        Bandpass
        Notch
    End Enum


    Private Const FilterTypeCount As Integer = FilterType.Notch + 1
    Private Const FMScaler As Integer = 10
    Private Const DriveScaler As Single = 5.0F
    Private Const DriveOffset As Single = 0.5F
    Private Const FrequencyMaxValue As Single = 12543.85F

    Private output1 As Single = 0
    Private output2 As Single = 0
    Private m_filterType As FilterType = FilterType.Lowpass
    Private frequency As Single = 0.5F
    Private resonance As Single = 0
    Private keyTracking As Single = 0
    Private frequencyNote As Integer
    Private keyTrackedNote As Integer = 0
    Private currentNote As Integer = 0
    Private input1 As MonoSynthComponent
    Private input2 As MonoSynthComponent
    Private m_fmModulator1 As MonoSynthComponent
    Private fmLevel1 As Single = 0
    Private m_fmModulator2 As MonoSynthComponent
    Private fmLevel2 As Single = 0
    Private m_amModulator As MonoSynthComponent
    Private drive As Single = DriveOffset
    Private m_synthesizeReplaceEnabled As Boolean = True



    Public Sub New(ByVal sampleRate As SampleRate, _
                   ByVal buffer As MonoBuffer, _
                   ByVal input1 As MonoSynthComponent, _
                   ByVal input2 As MonoSynthComponent, _
                   ByVal fmModulator1 As MonoSynthComponent, _
                   ByVal fmModulator2 As MonoSynthComponent, _
                    ByVal amModulator As MonoSynthComponent)

        MyBase.New(sampleRate, buffer)

        If input1 Is Nothing Then
            Throw New ArgumentNullException("input1")
        ElseIf input2 Is Nothing Then
            Throw New ArgumentNullException("input2")
        ElseIf fmModulator1 Is Nothing Then
            Throw New ArgumentNullException("fmModulator1")
        ElseIf fmModulator2 Is Nothing Then
            Throw New ArgumentNullException("fmModulator2")
        ElseIf amModulator Is Nothing Then
            Throw New ArgumentNullException("amModulator")
        End If

        Me.input1 = input1
        Me.input2 = input2
        Me.m_fmModulator1 = fmModulator1
        Me.m_fmModulator2 = fmModulator2
        Me.m_amModulator = amModulator

        SetParameterValue(ParameterId.CutoffFrequency, 0.5F)
    End Sub

    Public Sub New(ByVal sampleRate As SampleRate, _
                   ByVal buffer As MonoBuffer, _
                   ByVal name As String, _
                   ByVal input1 As MonoSynthComponent, _
                   ByVal input2 As MonoSynthComponent, _
                   ByVal fmModulator1 As MonoSynthComponent, _
                   ByVal fmModulator2 As MonoSynthComponent, _
                   ByVal amModulator As MonoSynthComponent)

        MyBase.New(sampleRate, buffer, name)

        If input1 Is Nothing Then
            Throw New ArgumentNullException("input1")
        ElseIf input2 Is Nothing Then
            Throw New ArgumentNullException("input2")
        ElseIf fmModulator1 Is Nothing Then
            Throw New ArgumentNullException("fmModulator1")
        ElseIf fmModulator2 Is Nothing Then
            Throw New ArgumentNullException("fmModulator2")
        ElseIf amModulator Is Nothing Then
            Throw New ArgumentNullException("amModulator")
        End If

        Me.input1 = input1
        Me.input2 = input2
        Me.m_fmModulator1 = fmModulator1
        Me.m_fmModulator2 = fmModulator2
        Me.m_amModulator = amModulator

        SetParameterValue(ParameterId.CutoffFrequency, 0.5F)
    End Sub



    Public Overloads Overrides Sub Synthesize(ByVal offset As Integer, ByVal count As Integer)
        Dim buffer As Single() = GetBuffer()
        Dim inputBuffer1 As Single() = input1.GetBuffer()
        Dim inputBuffer2 As Single() = input2.GetBuffer()
        Dim fmBuffer1 As Single() = m_fmModulator1.GetBuffer()
        Dim fmBuffer2 As Single() = m_fmModulator2.GetBuffer()
        Dim ampBuffer As Single() = m_amModulator.GetBuffer()
        Dim c As Integer = offset + count
        Dim fm As Single
        Dim modNote As Single
        Dim freq As Single
        Dim f As Single
        Dim fb As Single
        Dim input As Single
        Dim output As Single = 0

        Debug.Assert(inputBuffer1.Length = buffer.Length)
        Debug.Assert(fmBuffer1.Length = buffer.Length)
        Debug.Assert(fmBuffer2.Length = buffer.Length)
        Debug.Assert(ampBuffer.Length = buffer.Length)

        For i As Integer = offset To c - 1
            fm = fmBuffer1(i) * (fmLevel1 * fmLevel1) + fmBuffer2(i) * (fmLevel2 * fmLevel2)

            modNote = keyTrackedNote + NotesPerOctave * FMScaler * fm

            If modNote < 0 Then
                modNote = 0
            ElseIf modNote > NoteCount - 1 Then
                modNote = NoteCount - 1
            End If

            freq = PowerOfTwoTable.GetPower((modNote - A440NoteNumber) / NotesPerOctave) * A440Frequency
            f = freq / FrequencyMaxValue
            fb = resonance + resonance / (1.0F - f)

            input = (inputBuffer1(i) + inputBuffer2(i)) / 2 * (drive * drive)

            output1 += f * (input - output1 + fb * (output1 - output2))
            output2 += f * (output1 - output2)
            'If output1 > 100 Then output1 = 100
            'If output1 < -100 Then output1 = -100
            'If output2 > 100 Then output2 = 100
            'If output2 < -100 Then output2 = -100

            Select Case m_filterType
                Case FilterType.Lowpass
                    output = output2

                Case FilterType.Highpass
                    output = input - output2

                Case FilterType.Bandpass
                    output = output1 - output2

                Case FilterType.Notch
                    output = input - (output1 - output2)

                Case Else
                    Debug.Fail("Unhandled filter type.")

            End Select

            ' ----------------------------------------------------------------- only +0.6 to -0.6 (not +1/-1)
            If output >= 1 Then
                output = 0.66666F
            ElseIf output <= -1 Then
                output = -0.66666F
            Else
                output = output - output * output * output / 3.0F
            End If

            ' ----------------------------------------------------------------- TODO - best
            'output = 1.5F * output - 0.5F * output * output * output
            'If output > 1 Then output = 1
            'If output < -1 Then output = -1


            output *= ampBuffer(i)

            buffer(i) = (If(m_synthesizeReplaceEnabled, output, buffer(i) + output))
        Next
    End Sub

    Public Overloads Overrides Sub Trigger(ByVal previousNote As Integer, ByVal note As Integer, ByVal velocity As Single)
        currentNote = note
        CalculateKeyTrackedNote()
    End Sub

    Private Sub CalculateKeyTrackedNote()
        keyTrackedNote = CInt(Math.Round(frequencyNote + (currentNote - MiddleCNoteNumber) * keyTracking))

        If keyTrackedNote < 0 Then
            keyTrackedNote = 0
        ElseIf keyTrackedNote >= NoteCount Then
            keyTrackedNote = NoteCount - 1
        End If
    End Sub

    Public Overloads Overrides Sub Release(ByVal velocity As Single)
    End Sub


    Public Property FMModulator1() As MonoSynthComponent
        Get
            Return m_fmModulator1
        End Get
        Set(ByVal value As MonoSynthComponent)
            If value Is Nothing Then
                Throw New ArgumentNullException("value")
            End If

            Dim oldOrdinal As Integer = Ordinal

            m_fmModulator1 = value

            If oldOrdinal <> Ordinal Then
                OnOrdinalValueChanged(EventArgs.Empty)
            End If
        End Set
    End Property

    Public Property FMModulator2() As MonoSynthComponent
        Get
            Return m_fmModulator2
        End Get
        Set(ByVal value As MonoSynthComponent)
            If value Is Nothing Then
                Throw New ArgumentNullException("value")
            End If
            m_fmModulator2 = value
            Dim oldOrdinal As Integer = Ordinal
            If oldOrdinal <> Ordinal Then
                OnOrdinalValueChanged(EventArgs.Empty)
            End If
        End Set
    End Property

    Public Property AmModulator() As MonoSynthComponent
        Get
            Return m_amModulator
        End Get
        Set(ByVal value As MonoSynthComponent)
            If value Is Nothing Then
                Throw New ArgumentNullException("value")
            End If
            m_amModulator = value
            Dim oldOrdinal As Integer = Ordinal
            If oldOrdinal <> Ordinal Then
                OnOrdinalValueChanged(EventArgs.Empty)
            End If
        End Set
    End Property

    Public Overloads Overrides ReadOnly Property Ordinal() As Integer
        Get
            Return 1 + input1.Ordinal + input2.Ordinal + m_fmModulator1.Ordinal + m_fmModulator2.Ordinal + m_amModulator.Ordinal
        End Get
    End Property

    Public Overloads Overrides Property SynthesizeReplaceEnabled() As Boolean
        Get
            Return m_synthesizeReplaceEnabled
        End Get
        Set(ByVal value As Boolean)
            m_synthesizeReplaceEnabled = value
        End Set
    End Property


#Region "IProgrammable Members"

    Public Function GetParameterName(ByVal index As Integer) As String Implements IProgrammable.GetParameterName
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As String = String.Empty
        Dim name__1 As String = Name
        If Not String.IsNullOrEmpty(name__1) Then name__1 = name__1 & " "

        Select Case CType(index, ParameterId)
            Case ParameterId.CutoffFrequency
                result = name__1 & "Cutoff Frequency"
                
            Case ParameterId.FilterType
                result = name__1 & "Filter Type"
                
            Case ParameterId.Resonance
                result = name__1 & "Resonance"
                
            Case ParameterId.KeyboardTracking
                result = name__1 & "Keyboard Tracking"
                
            Case ParameterId.FMLevel1
                result = name__1 & "FM 1 Level"
                
            Case ParameterId.FMPolarity1
                result = name__1 & "FM 1 Polarity"
                
            Case ParameterId.FMLevel2
                result = name__1 & "FM 2 Level"
                
            Case ParameterId.FMPolarity2
                result = name__1 & "FM 2 Polarity"
                
            Case ParameterId.Drive
                result = name__1 & "Drive"

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select

        Return result
    End Function

    Public Function GetParameterLabel(ByVal index As Integer) As String Implements IProgrammable.GetParameterLabel
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As String = ""

        Select Case CType(index, ParameterId)
            Case ParameterId.CutoffFrequency
                result = "Hz"

            Case ParameterId.FilterType
                result = "Type"

            Case ParameterId.Resonance
                result = "Level"

            Case ParameterId.KeyboardTracking
                result = "Tracking"

            Case ParameterId.FMLevel1
                result = "dB"

            Case ParameterId.FMPolarity1
                result = "+/-"

            Case ParameterId.FMLevel2
                result = "dB"

            Case ParameterId.FMPolarity2
                result = "+/-"

            Case ParameterId.Drive
                result = "dB"

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select

        Return result
    End Function

    Public Function GetParameterDisplay(ByVal index As Integer) As String Implements IProgrammable.GetParameterDisplay
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As String = String.Empty

        Select Case CType(index, ParameterId)
            Case ParameterId.CutoffFrequency
                If True Then
                    Dim f As Single = PowerOfTwoTable.GetPower((frequencyNote - A440NoteNumber) \ NotesPerOctave) * A440Frequency

                    result = f.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.FilterType
                result = m_filterType.ToString()

            Case ParameterId.Resonance
                If True Then
                    Dim r As Integer = CInt(Math.Truncate(resonance * ParameterScaler))

                    result = r.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.KeyboardTracking
                If True Then
                    Dim kt As Integer = CInt(Math.Truncate(keyTracking * ParameterScaler))

                    result = kt.ToString(Synthesizer.FormatInfo)
                End If

            Case ParameterId.FMLevel1
                result = Synthesizer.DecibelToString(fmLevel1)

            Case ParameterId.FMPolarity1
                result = (If(fmLevel1 >= 0, "+", "-"))

            Case ParameterId.FMLevel2
                result = Synthesizer.DecibelToString(fmLevel2)

            Case ParameterId.FMPolarity2
                result = (If(fmLevel2 >= 0, "+", "-"))

            Case ParameterId.Drive
                result = Synthesizer.DecibelToString((drive - DriveOffset) / DriveScaler)

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select

        Return result
    End Function

    Public Function GetParameterValue(ByVal index As Integer) As Single Implements IProgrammable.GetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        End If

        Dim result As Single = 0

        Select Case CType(index, ParameterId)
            Case ParameterId.CutoffFrequency
                result = frequency

            Case ParameterId.FilterType
                result = CSng(m_filterType / (FilterTypeCount - 1))

            Case ParameterId.Resonance
                result = resonance

            Case ParameterId.KeyboardTracking
                result = keyTracking

            Case ParameterId.FMLevel1
                result = Math.Abs(fmLevel1)

            Case ParameterId.FMPolarity1
                result = (If(fmLevel1 >= 0, 0, 1))

            Case ParameterId.FMLevel2
                result = Math.Abs(fmLevel2)

            Case ParameterId.FMPolarity2
                result = (If(fmLevel2 >= 0, 0, 1))

            Case ParameterId.Drive
                result = (drive - DriveOffset) / DriveScaler

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select

        Return result
    End Function

    Public Sub SetParameterValue(ByVal index As Integer, ByVal value As Single) Implements IProgrammable.SetParameterValue
        If index < 0 OrElse index >= ParameterCount Then
            Throw New ArgumentOutOfRangeException("index")
        ElseIf value < 0 OrElse value > 1 Then
            Throw New ArgumentOutOfRangeException("value")
        End If

        Select Case CType(index, ParameterId)
            Case ParameterId.CutoffFrequency
                frequency = value
                frequencyNote = CInt(Math.Truncate((NoteCount - 1) * frequency))
                CalculateKeyTrackedNote()

            Case ParameterId.FilterType
                m_filterType = CType(Math.Round(value * (FilterTypeCount - 1)), FilterType)

            Case ParameterId.Resonance
                resonance = value

            Case ParameterId.KeyboardTracking
                keyTracking = value

            Case ParameterId.FMLevel1
                If True Then
                    Dim polarity As Single = (If(fmLevel1 < 0, -1, 1))

                    fmLevel1 = value * polarity
                End If

            Case ParameterId.FMPolarity1
                fmLevel1 = (If(value < 0.5, Math.Abs(fmLevel1), -Math.Abs(fmLevel1)))

            Case ParameterId.FMLevel2
                If True Then
                    Dim polarity As Single = (If(fmLevel2 < 0, -1, 1))

                    fmLevel2 = value * polarity
                End If

            Case ParameterId.FMPolarity2
                fmLevel2 = (If(value < 0.5, Math.Abs(fmLevel2), -Math.Abs(fmLevel2)))

            Case ParameterId.Drive
                drive = value * DriveScaler + DriveOffset

            Case Else
                Debug.Fail("Unhandled parameter")

        End Select
    End Sub

    Public ReadOnly Property ParameterCount() As Integer Implements IProgrammable.ParameterCount
        Get
            Return ParameterId.Drive + 1
        End Get
    End Property

#End Region

End Class
