

Imports System.Text

' ------------------------------------------------------------------------ base class for all synthesizer components
Public MustInherit Class SynthComponent
	Inherits Component

    Public Const NoteCount As Integer = 128
    Public Const NotesPerOctave As Integer = 12
    Public Const A440NoteNumber As Integer = 69
    Public Const A440Frequency As Single = 440
    Public Const MiddleCNoteNumber As Integer = 60
    Protected Const ParameterScaler As Integer = 127     ' For scaling generic parameter values.


    ' -------------------------------------------------- Occurs with the value of the Ordinal property changes.
    Public Event OrdinalChanged As EventHandler

    Protected Overridable Sub OnOrdinalValueChanged(ByVal e As EventArgs)
        RaiseEvent OrdinalChanged(Me, e)
    End Sub

    Protected Sub New(ByVal sampleRate As SampleRate)
        MyBase.New(sampleRate)
    End Sub

    Protected Sub New(ByVal sampleRate As SampleRate, ByVal name As String)
        MyBase.New(sampleRate, name)
    End Sub


    ' ------------------------------------------------ Synthesizes the SynthComponent's output into its buffer using 
    '                                                  the specified (zero-based) offset and count values 
    '                                                  to determine where to place the output in the buffer
    '
    Public MustOverride Sub Synthesize(ByVal offset As Integer, ByVal count As Integer)


    ' ------------------------------------------------ Triggers the SyntheComponent with the specified note
    '
    Public MustOverride Sub Trigger(ByVal previousNote As Integer, ByVal note As Integer, ByVal velocity As Single)

    ' ------------------------------------------------ Releases the currently triggered note
    '
    Public MustOverride Sub Release(ByVal velocity As Single)


    ' ------------------------------------------------ Gets or sets a value indicating whether the SynthComponent
    '                                                  should synthesize its output in-place or sum its output 
    '                                                  with the values already in its buffer
    '
    Public MustOverride Property SynthesizeReplaceEnabled() As Boolean


    ' ------------------------------------------------- Gets the SynthComponent's ordinal value.
    '
    Public MustOverride ReadOnly Property Ordinal() As Integer


    ' ------------------------------------------------- Gets or sets the SynthComponent's buffer size.
    '
    Friend MustOverride Property BufferSize() As Integer

End Class
