

Imports System.Collections
Imports System.Collections.Generic
Imports LC.Midi

' -------------------------------------------------------------- keeps track of which notes are currently playing
Friend NotInheritable Class NoteTracker
	Implements IEnumerable(Of ChannelMessage)

    ' The list of note-on messages currently playing.
    Private noteOnMessages As New List(Of ChannelMessage)()

    Private m_lastNote As Integer = 69


    ' -------------------------------------------------------------------- tracks a note-on messages
    '
    ' The NoteTracker will add the specified note-on message to its list
    ' of note-on messages. It will not remove the message until receiving
    ' a corresponding note-off message.
    '
    Public Sub NoteOn(ByVal message As ChannelMessage)
        If message Is Nothing Then
            Throw New ArgumentNullException("message")
        ElseIf message.Command <> ChannelCommand.NoteOn OrElse (message.Command = ChannelCommand.NoteOn AndAlso message.Data2 = 0) Then
            Throw New ArgumentException("Not a note on message.")
        End If
        noteOnMessages.Add(message)
        m_lastNote = message.Data1
    End Sub


    ' -------------------------------------------------------------------- tracks a note-off message
    '
    ' The NoteTracker will remove the note-on message that corresponds to
    ' the specified note-off message from its list of note-on messages,
    ' if one exists.
    '
    Public Sub NoteOff(ByVal message As ChannelMessage)
        If message Is Nothing Then
            Throw New ArgumentNullException("message")
        ElseIf message.Command <> ChannelCommand.NoteOff Then
            If message.Command = ChannelCommand.NoteOn Then
                If message.Data2 > 0 Then
                    Throw New ArgumentException("Not a note off message.")
                End If
            Else
                Throw New ArgumentException("Not a note off message.")
            End If
        End If

        ' Find the index into the list for the corresponding note-on message.
        'Dim index As Integer = noteOnMessages.FindIndex(AddressOf NoteOnIndexFinder)
        ' TODO - LIVIO - OK ???
        Dim index As Integer = -1
        For i As Int32 = 0 To noteOnMessages.Count - 1
            If message.MidiChannel = noteOnMessages(i).MidiChannel AndAlso message.Data1 = noteOnMessages(i).Data1 Then
                index = i
                Exit For
            End If
        Next

        ' If a corresponding note-on message was found.
        If index >= 0 Then
            ' Remove it from the list of note-on messages.
            noteOnMessages.RemoveAt(index)

            If noteOnMessages.Count > 0 Then
                m_lastNote = noteOnMessages(noteOnMessages.Count - 1).Data1
            End If
        End If
    End Sub

    ' -------------------------------------------------------------------- clears all note-on messages from the list
    Public Sub Clear()
        noteOnMessages.Clear()
    End Sub

    ' -------------------------------------------------------------------- number of note-on messages currently tracked
    Public ReadOnly Property Count() As Integer
        Get
            Return noteOnMessages.Count
        End Get
    End Property

    ' -------------------------------------------------------------------- the latest note currently playing 
    Public ReadOnly Property LastNote() As Integer
        Get
            Return m_lastNote
        End Get
    End Property


#Region "IEnumerable<ChannelMessage> Members"

    Public Function GetEnumerator() As IEnumerator(Of ChannelMessage) Implements IEnumerable(Of ChannelMessage).GetEnumerator
        Return noteOnMessages.GetEnumerator()
    End Function

#End Region

#Region "IEnumerable Members"

    Private Function IEnumerable_GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        Return noteOnMessages.GetEnumerator()
    End Function

#End Region

End Class
