

Imports System.Collections
Imports System.Collections.Generic
Imports System.IO

' ------------------------------------------------------------------------ a collection of Waveforms
Public Class Wavetable
    Implements IEnumerable(Of Waveform)

    ' --------------------------------------------------------------------- Wavetable cache. 
    '                                                                       Wavetables are stored at the class level
    '                                                                       and are reused.
    Private Shared m_wavetables As New Dictionary(Of String, Wavetable)()
    ' --------------------------------------------------------------------- waveform collection that make up the Wavetable 
    Private waveforms As IList(Of Waveform)


    Public Sub New(ByVal waveforms As IEnumerable(Of Waveform))
        Me.waveforms = New List(Of Waveform)(waveforms)
    End Sub

    ' ---------------------------------------------------------------------- clear the class level Wavetables
    Public Shared Sub Clear()
        m_wavetables.Clear()
    End Sub

    ' ---------------------------------------------------------------------- loads a Wavetable using the name
    Public Shared Function Load(ByVal name As String) As Wavetable
        If name Is Nothing Then
            Throw New ArgumentNullException("name")
        End If

        ' ------------------------------------------------------------------ if this Wavetable has already been loaded
        If m_wavetables.ContainsKey(name) Then
            ' -------------------------------------------------------------- return the Wavetable that was loaded
            Return m_wavetables(name)
        End If

        ' For holding the waveforms read from the file with the specified file name.
        Dim waveforms As New List(Of Waveform)()

        ' Get the Wavetable data.
        Dim buffer As Byte() = DirectCast(Wavetables.ResourceManager.GetObject(name), Byte())

        Using stream As New MemoryStream(buffer)
            ' Wrap stream in a BinaryReader so that we can read the waveform data.
            Dim reader As New BinaryReader(stream)

            ' Read the number of waveforms contained in the collection.
            Dim count As Integer = reader.ReadInt32()

            ' For every waveform in the Wavetable.
            For i As Integer = 0 To count - 1
                ' Read the length of the waveform.
                Dim length As Integer = reader.ReadInt32()

                ' Create an array for holding the waveform data.
                Dim waveData As Single() = New Single(length - 1) {}

                ' Read the samples for this waveform.
                For n As Integer = 0 To length - 1
                    waveData(n) = CSng(reader.ReadSingle())
                Next

                ' Create a Waveform object using the waveform data 
                ' and add it to the collection of waveforms.
                waveforms.Add(New Waveform(waveData))
            Next

            reader.Close()
        End Using

        ' Create a Wavetable using the Waveforms read from stream.
        Dim waveformCollection As New Wavetable(waveforms)

        ' Store the Wavetable; it will be reused if an attempt 
        ' is made to load the same Wavetable.
        m_wavetables.Add(name, waveformCollection)

        Return waveformCollection
    End Function

    ' -------------------------------------------------------------------- Gets the Waveform at the (zero based) index
    Default Public ReadOnly Property Item(ByVal index As Integer) As Waveform
        Get
            If index < 0 OrElse index >= Count Then
                Throw New IndexOutOfRangeException()
            End If
            Return waveforms(index)
        End Get
    End Property

    ' --------------------------------------------------------------------- get the Waveform count in the Wavetable
    Public ReadOnly Property Count() As Integer
        Get
            Return waveforms.Count
        End Get
    End Property


#Region "IEnumerable<Waveform> Members"

    Public Function GetEnumerator() As IEnumerator(Of Waveform) Implements IEnumerable(Of Waveform).GetEnumerator
        Return waveforms.GetEnumerator()
    End Function

#End Region

#Region "IEnumerable Members"

    Private Function IEnumerable_GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
        Return waveforms.GetEnumerator()
    End Function

#End Region

End Class
