﻿
' ENGLISH
' This program is also a collection of useful functions for creating Syntetizers and tinkering with MIDI
' The commented lines on "synth2" show how to create more independent Synthesizers, each with a different timbre.
' The commented lines on "midiInputDevice" show how to receive data from the MIDI

' ITALIANO
' Questo programma e' anche una collezione di funzioni utili per creare Synth e trafficare coi MIDI
' Le righe commentate sul "synth2" mostrano come creare piu' Synth indipendenti, ognuno con un timbro diverso.
' Le righe commentate sul "midiInputDevice" mostrano come ricevere dati dal MIDI

' MIDI-ID   NOTE   Frequency
' --------------------------
'     0     Do0
'    12     Do1    16,352 Hz
'    24     Do2    32,703 Hz
'    36     Do3    65,406 Hz
'    48     Do4    130,81 Hz
'    60     Do5    261,63 Hz  (central DO also called Do4 by musicians) 
'    72     Do6    523,25 Hz
'    84     Do7    1046,5 Hz
'    96     Do8    2093,0 Hz
'   108     Do9    4186,0 Hz
'   120     Do10   8372,0 Hz
'   127     Sol10

' Visible Keyboard range is: Do2 to Do8 


Imports LC.Synth
Imports LC.Midi

Public Class Form1

    Friend ChordFileName As String = ""
    Friend ConfigFileName As String = ""

    Private WithEvents mtimer As MilliTimer
    Private Tuner1 As TunerMeter
    Private Tuner2 As TunerMeter


    Private Sub Form1_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        'Me.Left = -9000
        ' ---------------------------------------------------------------- 
        InitPlatformData()
        Dim s As String
        s = IO.Path.GetFileNameWithoutExtension(Application.StartupPath)
        If s.ToLower = "bin" Then
            s = IO.Path.GetFileNameWithoutExtension(IO.Directory.GetParent(Application.StartupPath).ToString)
        End If
        Text = AppTitleAndVersion("Theremino System - Theremin")
        'If s <> "Theremino_Theremin" Then Me.Text &= "  ( " & s & " )"
        ' ----------------------------------------------------------------
        System.Threading.Thread.CurrentThread.Priority = System.Threading.ThreadPriority.Highest
        System.Diagnostics.Process.GetCurrentProcess().PriorityClass = ProcessPriorityClass.RealTime
        ' ---------------------------------------------------------------- 
        EventsAreEnabled = False
        ' ---------------------------------------------------------------- init Synthesizer
        synth1 = CreateSynthesizer(0, 2048, 44100)
        'synth2 = CreateSynthesizer(0, 2048, 44100)

        ' ********************************************************
        ' If Audio Out is not working than quit the app
        ' ********************************************************
        Try
            Dim dummy As Int32 = synth1.BufferSize
        Catch
            MsgBox("Can not create the Output Device", MsgBoxStyle.Critical)
            End
        End Try

        ' ---------------------------------------------------------------- load the default bank file if it exists
        If IO.File.Exists(DefaultBank) Then
            synth1.LoadBank(DefaultBank)
            'synth2.LoadBank(DefaultBank)
            synth1.SetSelectedProgramIndex(0)
            'synth2.SetSelectedProgramIndex(0)
        End If
        ' ---------------------------------------------------------------- load user options
        cmb_PlayEffects.SelectedIndex = 0
        cmb_PlayerType.SelectedIndex = 0
        Load_INI()
        StartThereminoHAL()
        SetAllParams()
        ' ---------------------------------------------------------------- set user interface
        ToolStrip1.Renderer = New ToolStripButtonRenderer
        synth1.SetBufferSize(BufferLen)
        InitSynthEditor(synth1)
        FillProgramList()
        ' ---------------------------------------------------------------- load tonality list from file
        Load_ChordFile(ChordFileName)
        ChordDirectionPositive = True
        ParseTonalityLine()
        ' ---------------------------------------------------------------- start midi
        ShowOutDeviceName()
        OpenOutputDevice()
        'InitMidiInput()
        ' ---------------------------------------------------------------- start all
        synth1.PlayStart()
        ' ---------------------------------------------------------------- start MilliTimer
        mtimer = New MilliTimer()
        mtimer.Interval = PollingTime
        mtimer.Start()
        ' ---------------------------------------------------------------- user interface timer
        Timer1.Interval = 16
        Timer1.Start()
        ' ---------------------------------------------------------------- 
        EventsAreEnabled = True
        ' ---------------------------------------------------------------- start UserInterface
        Refresh()
        Me.Opacity = 1
    End Sub

    Private Sub Form1_Shown(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Shown
        Tuner1 = New TunerMeter(pbox_Tuner1)
        Tuner2 = New TunerMeter(pbox_Tuner2)
    End Sub

    Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        CloseAllAndExit()
    End Sub

    Private Sub Form_LocationChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LocationChanged
        If Not EventsAreEnabled Then Exit Sub
        LimitFormPosition(Me)
    End Sub

    Private Sub CloseAllAndExit()
        ' ----------------------------------------- Save Bank
        If BankModifyed Then
            If MsgBox("Sound bank has been modifyed, save the default bank?", MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
                SaveDefaultBank()
            End If
        End If
        ' ----------------------------------------- Save
        'Save_ChordFile()
        Save_INI()
        ' ----------------------------------------- Close MIDI
        CloseOutputDevice()
        'CloseMidiInput()
        ' ----------------------------------------- Stop MilliTimer
        If mtimer IsNot Nothing Then mtimer.Stop()
        ' ----------------------------------------- Close All
        synth1.PlayStop()
        synth1.Dispose()
        'synth2.PlayStop()
        'synth2.Dispose()
        ' -----------------------------------------
        StopThereminoHAL()
    End Sub


    ' =========================================================================
    '   THEREMINO HAL - START STOP -
    ' =========================================================================
    Private HalProcess As Process = Nothing
    Private Sub StartThereminoHAL()
        Dim path As String = Application.StartupPath & "\Theremino_HAL.exe"
        If Not My.Computer.FileSystem.FileExists(path) Then
            path = Application.StartupPath & "\Theremino_HAL\Theremino_HAL.exe"
        End If
        If Not My.Computer.FileSystem.FileExists(path) Then Return
        Dim psi As New ProcessStartInfo
        psi.WorkingDirectory = IO.Path.GetDirectoryName(path)
        psi.FileName = path
        HalProcess = Process.Start(psi)
    End Sub
    Private Sub StopThereminoHAL()
        If HalProcess IsNot Nothing AndAlso Not HalProcess.HasExited Then
            Try
                HalProcess.CloseMainWindow()
            Finally
            End Try
        End If
    End Sub

    ' ===================================================================================
    '   MenuStrip and ToolStrip Gradients
    ' ===================================================================================
    Private Sub MenuStrip1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles MenuStrip1.Paint
        Dim bounds As New Rectangle(0, 0, _
                                    MenuStrip1.Width, MenuStrip1.Height)
        Dim brush As New Drawing2D.LinearGradientBrush(bounds, _
                                                       Color.FromArgb(230, 230, 230), _
                                                       Color.FromArgb(200, 200, 200), _
                                                       Drawing2D.LinearGradientMode.Horizontal)
        e.Graphics.FillRectangle(brush, bounds)
    End Sub
    Private Sub ToolStrip1_Paint(ByVal sender As System.Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles ToolStrip1.Paint
        Dim bounds As New Rectangle(0, 0, _
                                    ToolStrip1.Width, ToolStrip1.Height)
        Dim brush As New Drawing2D.LinearGradientBrush(bounds, _
                                                       Color.White, _
                                                       Color.FromArgb(200, 200, 200), _
                                                       Drawing2D.LinearGradientMode.Vertical)
        e.Graphics.FillRectangle(brush, bounds)
    End Sub

    ' ===================================================================================
    '   ToolStrip PressedButton color
    ' ===================================================================================
    Class ToolStripButtonRenderer
        Inherits System.Windows.Forms.ToolStripProfessionalRenderer
        Protected Overrides Sub OnRenderButtonBackground(ByVal e As ToolStripItemRenderEventArgs)
            Dim btn As ToolStripButton = CType(e.Item, ToolStripButton)
            If btn IsNot Nothing AndAlso btn.CheckOnClick AndAlso btn.Checked Then
                Dim bounds As Rectangle = New Rectangle(0, 0, e.Item.Width - 1, e.Item.Height - 1)
                Dim brush As New Drawing2D.LinearGradientBrush(bounds, _
                                                               Color.Gold, _
                                                               Color.FromArgb(250, 250, 250), _
                                                               Drawing2D.LinearGradientMode.Vertical)
                e.Graphics.FillRectangle(brush, bounds)
                e.Graphics.DrawRectangle(Pens.Orange, bounds)
            Else
                MyBase.OnRenderButtonBackground(e)
            End If
        End Sub
    End Class

    ' ===================================================================================
    '  MenuStrip and ToolStrip accepting the first click
    '  If the form receives a WM_PARENTNOTIFY (528) message and is not focused 
    '  then the form is activated before to exec the message
    ' ===================================================================================
    Protected Overrides Sub WndProc(ByRef m As Message)
        If m.Msg = 528 AndAlso Not Me.Focused Then
            Me.Activate()
        End If
        MyBase.WndProc(m)
    End Sub


    ' =======================================================================================
    '   MENU FILE
    ' =======================================================================================
    Private Sub Menu_File_LoadConfiguration_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_LoadConfiguration.Click
        EventsAreEnabled = False
        LoadConfiguration()
        EventsAreEnabled = True
    End Sub
    Private Sub Menu_File_SaveConfigurationAs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_SaveConfigurationAs.Click
        SaveConfigurationAs()
    End Sub
    Private Sub Menu_File_SaveDefaultBank_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_SaveDefaultBank.Click
        SaveDefaultBank()
    End Sub
    Private Sub Menu_File_LoadBank_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_LoadBank.Click
        LoadBank()
    End Sub
    Private Sub Menu_File_SaveBankAs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_SaveBankAs.Click
        SaveBankAs()
    End Sub
    Private Sub Menu_File_LoadSingleVoice_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_LoadSingleVoice.Click
        LoadSingleVoice()
    End Sub
    Private Sub Menu_File_SaveSingleVoiceAs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_SaveSingleVoiceAs.Click
        SaveSingleVoiceAs()
    End Sub
    Private Sub Menu_File_SaveImage_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_SaveImage.Click
        SaveImage("ThereminoMCA_", Me)
    End Sub
    Private Sub Menu_File_Exit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_Exit.Click
        CloseAllAndExit()
    End Sub


    ' =======================================================================================
    '   MENU HELP
    ' =======================================================================================
    Dim Language As String
    Private Sub Menu_Help_ProgramHelp_ENG_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Help_ProgramHelp_ENG.Click
        Language = "ENG"
        OpenLocalizedHelp("ThereminHelp", ".pdf")
    End Sub
    Private Sub Menu_Help_ProgramHelp_ITA_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Help_ProgramHelp_ITA.Click
        Language = "ITA"
        OpenLocalizedHelp("ThereminHelp", ".pdf")
    End Sub
    Private Sub Menu_Help_ProgramHelp_SPA_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Help_ProgramHelp_ESP.Click
        Language = "ESP"
        OpenLocalizedHelp("ThereminHelp", ".pdf")
    End Sub
    Private Sub Menu_Help_ProgramHelp_JAP_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Help_ProgramHelp_JAP.Click
        Language = "JAP"
        OpenLocalizedHelp("ThereminHelp", ".pdf")
    End Sub
    Private Sub ToDoListToolStripMenuItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToDoListToolStripMenuItem.Click
        OpenLocalizedHelp("ToDo", ".txt")
    End Sub
    Private Sub OpenLocalizedHelp(ByVal name As String, Optional ByVal ext As String = ".rtf")

        Dim fname As String = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & "_" & Strings.Left(Language, 3).ToUpper & ext)
        If FileExists(fname) Then
            Process.Start(fname)
        Else
            fname = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & "_ENG" & ext)
            If FileExists(fname) Then
                Process.Start(fname)
            Else
                fname = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & "_ITA" & ext)
                If FileExists(fname) Then
                    Process.Start(fname)
                Else
                    fname = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & ext)
                    If FileExists(fname) Then
                        Process.Start(fname)
                    End If
                End If
            End If
        End If
    End Sub
    Private Sub Menu_Help_OpenProgramFolder_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Help_OpenProgramFolder.Click
        Process.Start(Application.StartupPath)
    End Sub

    ' =======================================================================================
    '   MENU ABOUT
    ' =======================================================================================
    Private Sub Menu_About_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_About.Click
        Form_About.ShowDialog()
    End Sub


    ' =======================================================================================
    '   TOOLBAR
    ' =======================================================================================
    Private Sub ToolStripButton_LoadConfiguration_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_LoadConfiguration.Click
        EventsAreEnabled = False
        LoadConfiguration()
        Load_ChordFile(ChordFileName)
        If ListBox_Chords.Items.Count > 0 Then
            ListBox_Chords.SelectedIndex = 0
            ParseTonalityLine()
        End If
        EventsAreEnabled = True
    End Sub
    Private Sub ToolStripButton_SaveConfiguration_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_SaveConfigurationAs.Click
        SaveConfigurationAs()
    End Sub
    Private Sub ToolStripButton_SaveDefaultBank_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_SaveDefaultBank.Click
        SaveDefaultBank()
    End Sub
    Private Sub ToolStripButton_LoadBank_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_LoadBank.Click
        LoadBank()
    End Sub
    Private Sub ToolStripButton_SaveBankAs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_SaveBankAs.Click
        SaveBankAs()
    End Sub
    Private Sub ToolStripButton_LoadVoice_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_LoadVoice.Click
        LoadSingleVoice()
    End Sub
    Private Sub ToolStripButton_SaveVoiceAs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_SaveVoiceAs.Click
        SaveSingleVoiceAs()
    End Sub
    Private Sub ToolStripButton_Run_CheckedChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles ToolStripButton_Run.CheckedChanged
        If Not EventsAreEnabled Then Return
        If ToolStripButton_Run.Checked Then
            '
        Else
            PlayingNotes_ReleaseAllNotes()
        End If
        ResetAllSlots()
    End Sub


    ' =======================================================================================
    '   PROPERTIES
    ' =======================================================================================
    Private Sub Controls_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles txt_NoteSlot.LostFocus, _
                                                                                                txt_TriggerLevel.LostFocus, _
                                                                                                txt_VolumeSlot.LostFocus, _
                                                                                                chk_VolumeNegative.LostFocus, _
                                                                                                cmb_PlayEffects.LostFocus, _
                                                                                                txt_Amount.LostFocus, _
                                                                                                cmb_PlayerType.LostFocus, _
                                                                                                txt_PollingTime.LostFocus, _
                                                                                                txt_BufferLen.LostFocus, _
                                                                                                txt_VoiceSlotPlus.LostFocus, _
                                                                                                txt_VoiceSlotMinus.LostFocus, _
                                                                                                txt_ChordSlotPlus.LostFocus, _
                                                                                                txt_ChordSlotMinus.LostFocus, _
                                                                                                txt_ChordTime.LostFocus, _
                                                                                                txt_OutMidiChannel.LostFocus
        If Not EventsAreEnabled Then Exit Sub
        Save_INI()
    End Sub

    Private Sub Controls_Changed(ByVal sender As Object, ByVal e As System.EventArgs) Handles txt_NoteSlot.TextChanged, _
                                                                                              txt_TriggerLevel.TextChanged, _
                                                                                              txt_VolumeSlot.TextChanged, _
                                                                                              chk_VolumeNegative.CheckedChanged, _
                                                                                              txt_Amount.TextChanged, _
                                                                                              txt_PollingTime.TextChanged, _
                                                                                              txt_BufferLen.TextChanged, _
                                                                                              txt_VoiceSlotPlus.TextChanged, _
                                                                                              txt_VoiceSlotMinus.TextChanged, _
                                                                                              txt_ChordSlotPlus.TextChanged, _
                                                                                              txt_ChordSlotMinus.TextChanged, _
                                                                                              txt_ChordTime.TextChanged, _
                                                                                              txt_OutMidiChannel.TextChanged
        SetAllParams()
    End Sub

    Private NoteSlot As Int32
    Private TriggerLevel As Int32
    Private VolumeSlot As Int32
    Private VolumeNegative As Boolean
    Private Amount As Int32
    Private PollingTime As Int32
    Private BufferLen As Int32
    Private VoiceSlotPlus As Int32
    Private VoiceSlotMinus As Int32
    Private ChordSlotPlus As Int32
    Private ChordSlotMinus As Int32
    Private ChordTime As Int32
    Private OutMidiChannel As Int32

    Private Sub SetAllParams()
        NoteSlot = txt_NoteSlot.NumericValueInteger
        TriggerLevel = txt_TriggerLevel.NumericValueInteger
        VolumeSlot = txt_VolumeSlot.NumericValueInteger
        VolumeNegative = chk_VolumeNegative.Checked
        Amount = txt_Amount.NumericValueInteger
        PollingTime = txt_PollingTime.NumericValueInteger
        BufferLen = txt_BufferLen.NumericValueInteger
        VoiceSlotPlus = txt_VoiceSlotPlus.NumericValueInteger
        VoiceSlotMinus = txt_VoiceSlotMinus.NumericValueInteger
        ChordSlotPlus = txt_ChordSlotPlus.NumericValueInteger
        ChordSlotMinus = txt_ChordSlotMinus.NumericValueInteger
        ChordTime = txt_ChordTime.NumericValueInteger
        OutMidiChannel = txt_OutMidiChannel.NumericValueInteger
    End Sub

    Private Sub txt_BufferLen_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txt_BufferLen.TextChanged
        If Not EventsAreEnabled Then Exit Sub
        txt_BufferLen.Refresh()
        synth1.SetBufferSize(txt_BufferLen.NumericValueInteger)
    End Sub

    Private Sub txt_PollingTime_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txt_PollingTime.TextChanged
        If Not EventsAreEnabled Then Exit Sub
        mtimer.Interval = txt_PollingTime.NumericValueInteger
    End Sub


    ' ====================================================================================================
    '  LOAD SAVE - VOICE BANK
    ' ====================================================================================================
    Private DefaultBank As String = PlatformAdjustedFileName(Application.StartupPath & "\Media\VoiceBanks\DefaultBank.bnk")

    Private Sub SaveDefaultBank()
        synth1.SaveBank(DefaultBank)
        BankModifyed = False
        Menu_File_SaveDefaultBank.Enabled = False
        ToolStripButton_SaveDefaultBank.Enabled = False
    End Sub
    Private Sub LoadBank()
        Dim ofd As OpenFileDialog = New OpenFileDialog
        ofd.InitialDirectory = PlatformAdjustedFileName(Application.StartupPath & "\Media\VoiceBanks\")
        ofd.Filter = "Bank file (*.bnk)|*bnk"
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            synth1.LoadBank(ofd.FileName)
            FillProgramList()
            Set_SoundBankModifyed_Flag()
        End If
    End Sub
    Private Sub SaveBankAs()
        Dim sfd As SaveFileDialog = New SaveFileDialog()
        sfd.InitialDirectory = PlatformAdjustedFileName(Application.StartupPath & "\Media\VoiceBanks\")
        sfd.DefaultExt = ".bnk"
        sfd.Filter = "Bank file (*.bnk)|*bnk"
        sfd.FileName = "Bank_" & Date.Now.ToString("yyyy_MM_dd_HH_mm_ss")
        If sfd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            synth1.SaveBank(sfd.FileName)
        End If
    End Sub
    Private Sub Set_SoundBankModifyed_Flag()
        BankModifyed = True
        Menu_File_SaveDefaultBank.Enabled = True
        ToolStripButton_SaveDefaultBank.Enabled = True
    End Sub

    ' ====================================================================================================
    '   LOAD SAVE - SINGLE VOICE
    ' ====================================================================================================
    Private Sub LoadSingleVoice()
        Dim ofd As OpenFileDialog = New OpenFileDialog
        ofd.InitialDirectory = PlatformAdjustedFileName(Application.StartupPath & "\Media\Voices\")
        ofd.Filter = "Voice file (*.voice)|*voice"
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            synth1.LoadProgram(ofd.FileName)
            ofd.FileName = IO.Path.GetFileNameWithoutExtension(ofd.FileName)
            Dim i As Int32 = synth1.SelectedProgramIndex
            ListBox_Programs.Items(i) = ofd.FileName
            synth1.SelectedProgramName = ofd.FileName
            Set_SoundBankModifyed_Flag()
        End If
    End Sub
    Private Sub SaveSingleVoiceAs()
        Dim sfd As SaveFileDialog = New SaveFileDialog()
        sfd.InitialDirectory = PlatformAdjustedFileName(Application.StartupPath & "\Media\Voices\")
        sfd.DefaultExt = ".voice"
        sfd.Filter = "Voice file (*.voice)|*voice"
        sfd.FileName = synth1.SelectedProgramName
        If sfd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            synth1.SaveProgram(sfd.FileName)
        End If
    End Sub


    ' ====================================================================================================
    '   RENAME and UP-DOWN - SINGLE VOICE
    ' ====================================================================================================
    Private Sub ButtonRename_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ButtonRename.Click
        Dim i As Int32 = synth1.SelectedProgramIndex
        Dim newName As String = InputBox("Select a new name." & vbCr & vbCr & "Max 24 chars", _
                                         "Rename voice", _
                                         synth1.SelectedProgramName, _
                                         Me.Left + 274, _
                                         Me.Top + 120)
        If newName = "" Then Return
        newName = Strings.Left(newName, 24).Trim
        synth1.SelectedProgramName = newName
        ListBox_Programs.Items(i) = newName
        Set_SoundBankModifyed_Flag()
    End Sub
    Private Sub ButtonUp_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ButtonUp.Click
        Dim i As Int32 = synth1.SelectedProgramIndex
        If i <= 0 Then Return
        ' ------------------------------------------------------------- exchange programs
        Dim filePath As String
        filePath = PlatformAdjustedFileName(Application.StartupPath & "\Media\Voices\")
        Dim file1 As String = filePath & "_______________file1"
        Dim file2 As String = filePath & "_______________file2"
        synth1.SaveProgram(file2)
        synth1.SetSelectedProgramIndex(i - 1)
        synth1.SaveProgram(file1)
        synth1.LoadProgram(file2)
        synth1.SetSelectedProgramIndex(i)
        synth1.LoadProgram(file1)
        IO.File.Delete(file1)
        IO.File.Delete(file2)
        ' ------------------------------------------------------------- exchange list names
        Dim dummy As String = ListBox_Programs.Items(i).ToString
        ListBox_Programs.Items(i) = ListBox_Programs.Items(i - 1)
        ListBox_Programs.Items(i - 1) = dummy
        ' ------------------------------------------------------------- 
        ListBox_Programs.SelectedIndex = i - 1
        synth1.SetSelectedProgramIndex(i - 1)
        Set_SoundBankModifyed_Flag()
    End Sub
    Private Sub ButtonDw_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ButtonDw.Click
        Dim i As Int32 = synth1.SelectedProgramIndex
        If i >= synth1.ProgramCount - 1 Then Return
        ' ------------------------------------------------------------- exchange programs
        Dim filePath As String
        filePath = PlatformAdjustedFileName(Application.StartupPath & "\Media\Voices\")
        Dim file1 As String = filePath & "_______________file1"
        Dim file2 As String = filePath & "_______________file2"
        synth1.SaveProgram(file1)
        synth1.SetSelectedProgramIndex(i + 1)
        synth1.SaveProgram(file2)
        synth1.LoadProgram(file1)
        synth1.SetSelectedProgramIndex(i)
        synth1.LoadProgram(file2)
        IO.File.Delete(file1)
        IO.File.Delete(file2)
        ' ------------------------------------------------------------- exchange list names
        Dim dummy As String = ListBox_Programs.Items(i).ToString
        ListBox_Programs.Items(i) = ListBox_Programs.Items(i + 1)
        ListBox_Programs.Items(i + 1) = dummy
        ' -------------------------------------------------------------
        ListBox_Programs.SelectedIndex = i + 1
        synth1.SetSelectedProgramIndex(i + 1)
        Set_SoundBankModifyed_Flag()
    End Sub


    ' ====================================================================================================
    '  CHORDS and TONALITY commands
    ' ====================================================================================================
    Private Sub btn_EditChords_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btn_EditChords.Click
        Process.Start(Application.StartupPath & "\media\chords\")
    End Sub

    Private Sub btn_LoadChords_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btn_LoadChords.Click
        ChordFileName = PlatformAdjustedFileName(ChordFileName, Application.StartupPath & "\media\chords\default.txt")
        '
        Dim ofd As OpenFileDialog = New OpenFileDialog
        If IO.File.Exists(ChordFileName) Then
            ofd.InitialDirectory = IO.Path.GetDirectoryName(ChordFileName)
            ofd.FileName = ChordFileName
        Else
            ofd.InitialDirectory = PlatformAdjustedFileName(Application.StartupPath & "\media\chords")
            ofd.FileName = ""
        End If
        ChordFileName = PlatformAdjustedFileName(ChordFileName)
        ofd.ShowDialog()
        ChordFileName = IO.Path.GetFileNameWithoutExtension(ofd.FileName)
        Load_ChordFile(ChordFileName)
        ListBox_Chords.SelectedIndex = 0
        ParseTonalityLine()
        Save_INI()
    End Sub

    Private ChordDirectionPositive As Boolean

    Delegate Sub NextChordDelegate()
    Private Sub NextChord()
        ChordDirectionPositive = True
        If ListBox_Chords.Items.Count = 0 Then Return
        Dim i As Int32 = ListBox_Chords.SelectedIndex + 1
        If i >= ListBox_Chords.Items.Count Then i = 0
        ListBox_Chords.SelectedIndex = i
    End Sub
    Private Sub NextChord_ThreadSafe()
        If ListBox_Chords.InvokeRequired Then
            Me.Invoke(New NextChordDelegate(AddressOf NextChord))
        Else
            NextChord()
        End If
    End Sub

    Delegate Sub LastChordDelegate()
    Private Sub LastChord()
        ChordDirectionPositive = False
        If ListBox_Chords.Items.Count = 0 Then Return
        Dim i As Int32 = ListBox_Chords.SelectedIndex - 1
        If i < 0 Then i = ListBox_Chords.Items.Count - 1
        ListBox_Chords.SelectedIndex = i
    End Sub
    Private Sub LastChord_ThreadSafe()
        If ListBox_Chords.InvokeRequired Then
            Me.Invoke(New LastChordDelegate(AddressOf LastChord))
        Else
            LastChord()
        End If
    End Sub

    Private ChordToBeSelected As Int32
    Delegate Sub SelectChordDelegate()
    Private Sub SelectChord()
        If ChordToBeSelected >= 0 And ChordToBeSelected < ListBox_Chords.Items.Count Then
            ListBox_Chords.SelectedIndex = ChordToBeSelected
        End If
    End Sub
    Private Sub SelectChord_ThreadSafe()
        If ListBox_Chords.InvokeRequired Then
            Me.Invoke(New SelectChordDelegate(AddressOf SelectChord))
        Else
            SelectChord()
        End If
    End Sub


    Private Sub ListBox_Tonality_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ListBox_Chords.SelectedIndexChanged
        If Not EventsAreEnabled Then Exit Sub
        ParseTonalityLine()
    End Sub

    Private RecursionLevel As Int32
    Private Sub ParseTonalityLine()
        '
        PlayingNotes_ReleaseAllNotes()
        '
        Dim i As Int32 = ListBox_Chords.SelectedIndex
        If i < 0 Then Return
        Dim l As String = ListBox_Chords.Items(i).ToString.ToUpper
        Dim c As Chords = Chords.InvalidChord
        '
        Dim nCharsDecoded As Int32 = 5
        Select Case Strings.Left(l, 5)
            Case "RANGE"
                l = Strings.Mid(l, nCharsDecoded + 1).Trim
                ' ----------------------------------------- Parse Note Range
                ParseNoteRange(l)
            Case Else
                ' ----------------------------------------- Parse BaseNote
                Dim BaseNote As Int32 = ParseNotes(l)
                ' ----------------------------------------- Parse Chord
                c = ParseChords(l)
                ' ----------------------------------------- SetBaseNoteAndChord
                Synth_SetBaseNoteAndChord(BaseNote, c)
        End Select
        ' ------------------------------------------------- if possible goto the firs valid chord
        If c = Chords.InvalidChord Then
            RecursionLevel += 1
            If RecursionLevel < ListBox_Chords.Items.Count Then
                If ChordDirectionPositive Then
                    NextChord_ThreadSafe()
                Else
                    LastChord_ThreadSafe()
                End If
                ParseTonalityLine()
            End If
        Else
            RecursionLevel = 0
        End If
    End Sub

    Private Sub ParseNoteRange(ByRef l As String)
        FirstNote = ParseNotes(l)
        FirstNote += 12 * CInt(Val(Strings.Left(l, 1)))
        l = Strings.Mid(l, 2).Trim
        LastNote = ParseNotes(l)
        LastNote += 12 * CInt(Val(l))
    End Sub

    Private Function ParseNotes(ByRef l As String) As Int32
        Dim n As Int32 = 0
        Dim nCharsDecoded As Int32 = 4
        Select Case Strings.Left(l, 4)
            Case "SOL#" : n = 8
            Case Else
                nCharsDecoded = 3
                Select Case Strings.Left(l, 3)
                    Case "ALL" : n = 0 : nCharsDecoded = 0
                    Case "WHI" : n = 0 : nCharsDecoded = 0
                    Case "SEV" : n = 0 : nCharsDecoded = 0
                    Case "DO#" : n = 1
                    Case "RE#" : n = 3
                    Case "FA#" : n = 6
                    Case "SOL" : n = 7
                    Case "LA#" : n = 10
                    Case Else
                        nCharsDecoded = 2
                        Select Case Strings.Left(l, 2)
                            Case "DO" : n = 0
                            Case "C#" : n = 1
                            Case "RE" : n = 2
                            Case "D#" : n = 3
                            Case "MI" : n = 4
                            Case "FA" : n = 5
                            Case "F#" : n = 6
                            Case "G#" : n = 8
                            Case "LA" : n = 9
                            Case "A#" : n = 10
                            Case "SI" : n = 11
                            Case Else
                                nCharsDecoded = 1
                                Select Case Strings.Left(l, 1)
                                    Case "C" : n = 0
                                    Case "D" : n = 2
                                    Case "E" : n = 4
                                    Case "F" : n = 5
                                    Case "G" : n = 7
                                    Case "A" : n = 9
                                    Case "B" : n = 11
                                    Case Else
                                        nCharsDecoded = 0
                                End Select
                        End Select
                End Select
        End Select
        l = Strings.Mid(l, nCharsDecoded + 1).Trim
        Return n
    End Function

    Private Function ParseChords(ByRef l As String) As Chords
        Dim c As Chords = Chords.InvalidChord
        Select Case l
            Case "MAJ", "M", "" : c = Chords.Major
            Case "6", "MAJ6" : c = Chords.Major6
            Case "7", "MAJ7" : c = Chords.Major7
            Case "7+", "MAJ7+" : c = Chords.Major7plus
            Case "9", "MAJ9" : c = Chords.Major9
            Case "9+", "MAJ9+" : c = Chords.Major9plus
            Case "69", "MAJ69", "MAJ6/9" : c = Chords.Major69
            Case "79", "MAJ79", "MAJ7/9" : c = Chords.Major79
            Case "11", "MAJ11" : c = Chords.Major11
            Case "11+", "MAJ11+" : c = Chords.Major11plus
            Case "13", "MAJ13" : c = Chords.Major13
            Case "13+", "MAJ13+" : c = Chords.Major13plus

            Case "MIN", "-" : c = Chords.Minor
            Case "MIN6", "-6" : c = Chords.Minor6
            Case "MIN7", "-7" : c = Chords.Minor7
            Case "MIN7+", "-7+" : c = Chords.Minor7plus
            Case "MIN9", "-9" : c = Chords.Minor9
            Case "MIN9+", "-9+" : c = Chords.Minor9plus
            Case "MIN69", "-69" : c = Chords.Minor69
            Case "MIN79", "-79" : c = Chords.Minor79
            Case "MIN11" : c = Chords.Minor11
            Case "MIN11+" : c = Chords.Minor11plus
            Case "MIN13" : c = Chords.Minor13
            Case "MIN13+" : c = Chords.Minor13plus

            Case "AUG" : c = Chords.Augmented
            Case "AUG6" : c = Chords.Augmented6
            Case "AUG7" : c = Chords.Augmented7
            Case "AUG7+" : c = Chords.Augmented7plus
            Case "AUG9" : c = Chords.Augmented9
            Case "AUG9+" : c = Chords.Augmented9plus
            Case "AUG69" : c = Chords.Augmented69
            Case "AUG79" : c = Chords.Augmented79

            Case "DIM" : c = Chords.Diminished
            Case "DIM6" : c = Chords.Diminished6
            Case "DIM7" : c = Chords.Diminished7
            Case "DIM7+" : c = Chords.Diminished7plus
            Case "DIM9" : c = Chords.Diminished9
            Case "DIM9+" : c = Chords.Diminished9plus
            Case "DIM69" : c = Chords.Diminished69
            Case "DIM79" : c = Chords.Diminished79

            Case "WHITE", "SEVENNOTES", "SEVEN NOTES" : c = Chords.SevenNotes
            Case "ALL", "ALLNOTES", "ALL NOTES" : c = Chords.AllNotes
        End Select
        Return c
    End Function



    ' ====================================================================================================
    '  BaseNote and Chords
    ' ====================================================================================================
    Private FirstNote As Int32
    Private LastNote As Int32
    Private EnabledNotes(127) As Int32
    Private EnabledNotesCount As Int32

    Friend Enum Chords
        Sus2
        Sus4
        '
        Major
        Major6
        Major7
        Major7plus
        Major9
        Major9plus
        Major69
        Major79
        Major11
        Major11plus
        Major13
        Major13plus
        '
        Minor
        Minor6
        Minor7
        Minor7plus
        Minor9
        Minor9plus
        Minor69
        Minor79
        Minor11
        Minor11plus
        Minor13
        Minor13plus
        '
        Augmented
        Augmented6
        Augmented7
        Augmented7plus
        Augmented9
        Augmented9plus
        Augmented69
        Augmented79
        '
        Diminished
        Diminished6
        Diminished7
        Diminished7plus
        Diminished9
        Diminished9plus
        Diminished69
        Diminished79
        '
        SevenNotes
        AllNotes
        '
        InvalidChord
    End Enum
    Private Sub Synth_SetBaseNoteAndChord(ByVal BaseNote As Int32, ByVal Chord As Chords)
        '
        Dim notes(-1) As Int32
        Select Case Chord
            Case Chords.Sus2 : notes = New Int32() {0, 2, 7}
            Case Chords.Sus4 : notes = New Int32() {0, 5, 7}
                '
            Case Chords.Major : notes = New Int32() {0, 4, 7}
            Case Chords.Major6 : notes = New Int32() {0, 4, 7, 9}
            Case Chords.Major7 : notes = New Int32() {0, 4, 7, 10}
            Case Chords.Major7plus : notes = New Int32() {0, 4, 7, 11}
            Case Chords.Major9 : notes = New Int32() {0, 2, 4, 7, 10}
            Case Chords.Major9plus : notes = New Int32() {0, 2, 4, 7, 11}
            Case Chords.Major69 : notes = New Int32() {0, 2, 4, 7, 9}
            Case Chords.Major79 : notes = New Int32() {0, 2, 4, 7, 10}
            Case Chords.Major11 : notes = New Int32() {0, 2, 4, 5, 7, 10}
            Case Chords.Major11plus : notes = New Int32() {0, 2, 4, 5, 7, 11}
            Case Chords.Major13 : notes = New Int32() {0, 2, 4, 7, 9, 10}
            Case Chords.Major13plus : notes = New Int32() {0, 2, 4, 7, 9, 11}
                '
            Case Chords.Minor : notes = New Int32() {0, 3, 7}
            Case Chords.Minor6 : notes = New Int32() {0, 3, 7, 9}
            Case Chords.Minor7 : notes = New Int32() {0, 3, 7, 10}
            Case Chords.Minor7plus : notes = New Int32() {0, 3, 7, 11}
            Case Chords.Minor9 : notes = New Int32() {0, 2, 3, 7, 10}
            Case Chords.Minor9plus : notes = New Int32() {0, 2, 3, 7, 11}
            Case Chords.Minor69 : notes = New Int32() {0, 2, 3, 7, 9}
            Case Chords.Minor79 : notes = New Int32() {0, 2, 3, 7, 10}
            Case Chords.Minor11 : notes = New Int32() {0, 2, 3, 5, 7, 10}
            Case Chords.Minor11plus : notes = New Int32() {0, 2, 3, 5, 7, 11}
            Case Chords.Minor13 : notes = New Int32() {0, 2, 3, 7, 9, 10}
            Case Chords.Minor13plus : notes = New Int32() {0, 2, 3, 7, 9, 11}
                '
            Case Chords.Augmented : notes = New Int32() {0, 4, 8}
            Case Chords.Augmented6 : notes = New Int32() {0, 4, 8, 9}
            Case Chords.Augmented7 : notes = New Int32() {0, 4, 8, 10}
            Case Chords.Augmented7plus : notes = New Int32() {0, 4, 8, 11}
            Case Chords.Augmented9 : notes = New Int32() {0, 2, 4, 8, 10}
            Case Chords.Augmented9plus : notes = New Int32() {0, 2, 4, 8, 11}
            Case Chords.Augmented69 : notes = New Int32() {0, 2, 4, 8, 9}
            Case Chords.Augmented79 : notes = New Int32() {0, 2, 4, 8, 10}
                '
            Case Chords.Diminished : notes = New Int32() {0, 3, 6}
            Case Chords.Diminished6 : notes = New Int32() {0, 3, 6, 9}
            Case Chords.Diminished7 : notes = New Int32() {0, 3, 6, 10}
            Case Chords.Diminished7plus : notes = New Int32() {0, 3, 6, 11}
            Case Chords.Diminished9 : notes = New Int32() {0, 2, 3, 6, 10}
            Case Chords.Diminished9plus : notes = New Int32() {0, 2, 3, 6, 11}
            Case Chords.Diminished69 : notes = New Int32() {0, 2, 3, 6, 9}
            Case Chords.Diminished79 : notes = New Int32() {0, 2, 3, 6, 10}
                '
            Case Chords.SevenNotes : notes = New Int32() {0, 2, 4, 5, 7, 9, 11}
            Case Chords.AllNotes : notes = New Int32() {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11}
        End Select
        '
        ' -------------------------------------------------------------------- set enabled notes from chord
        EnabledNotesCount = 0
        For i As Int32 = FirstNote + BaseNote To LastNote + BaseNote
            For j As Int32 = 0 To notes.Length - 1
                If i Mod 12 = (BaseNote + notes(j)) Mod 12 Then
                    EnabledNotes(EnabledNotesCount) = i
                    EnabledNotesCount += 1
                    Exit For
                End If
            Next
        Next
    End Sub

    Private Function CalcNoteID(ByVal NormalizedValue As Single) As Int32
        If Single.IsNaN(NormalizedValue) Then Return 0
        '
        Dim NoteID As Int32 = CInt(NormalizedValue * (EnabledNotesCount - 1))  ' 0(Do0) to 127(Fa#10)
        '
        If NoteID < 0 Then Return -1
        '
        Return EnabledNotes(NoteID)
    End Function


    ' ====================================================================================================
    '  Theremin Pizzicato
    ' ====================================================================================================
    Private OldNoteID As Int32 = -1

    Private Sub ThereminPlayer_PlayPizzicato(ByVal volume As Single, ByVal NormalizedValue As Single)
        If synth1 Is Nothing Then Return
        If Single.IsNaN(NormalizedValue) Then Return

        Dim NoteID As Int32 = CalcNoteID(NormalizedValue)
        If NoteID < 0 Then Return
        TunerNote = NoteID

        Dim Velocity As Int32 = CInt(volume * 127) ' translate volume(0..1) to velocity(0..127)

        If NoteID <> OldNoteID Then 'Or volume < 0.01 Then
            If OldNoteID >= 0 Then
                PlayingNotes_Release(OldNoteID)
            End If
            PlayingNotes_Retrigger(NoteID, Velocity)

            OldNoteID = NoteID
        Else
            PlayingNotes_Retrigger(NoteID, Velocity)
        End If
    End Sub


    ' ====================================================================================================
    '  MIXED FUNCTIONS
    ' ====================================================================================================
    Private Sub SendBendMessage(ByVal bend As Single)
        ' -----------------------------------------------
        '  Bend value
        ' ------------------------------------------------------------
        ' The two bytes combined together form a 14-bit value. 
        ' First byte - bits 0 to 6 = bits 0 to 6 of the 14-bit value 
        ' Second byte - bits 0 to 6 = bits 7 to 13 of the 14-bit value
        ' ------------------------------------------------------------
        Dim b As Int32 = CInt(16383 * bend)
        Dim m As ChannelMessage = New ChannelMessage(ChannelCommand.PitchWheel, _
                                                     0, _
                                                     b And 127, _
                                                     b >> 7)
        synth1.ProcessChannelMessage(m)
        MidiOutSendMessage(m)
    End Sub
    Private Sub ThereminPlayer_Stop()
        If OldNoteID >= 0 Then
            Dim Channel As Int32 = 0
            Dim message As ChannelMessage = New ChannelMessage(ChannelCommand.NoteOff, Channel, OldNoteID, 0)
            synth1.ProcessChannelMessage(message)
            MidiOutSendMessage(message)
            pianoControl1.ReleasePianoKey(OldNoteID)
            OldNoteID = -1
        End If
    End Sub
    Private Sub ResetAllSlots()
        For n As Int32 = 0 To 999
            Slots.WriteSlot(n, 0)
        Next
    End Sub

    ' ====================================================================================================
    '  COMBO PLAY EFFECTS
    ' ====================================================================================================
    Dim PlayEffect As Int32
    Private Sub cmb_PlayEffects_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmb_PlayEffects.SelectedIndexChanged
        PlayEffect = cmb_PlayEffects.SelectedIndex
    End Sub
    Private Sub cmb_PlayEffects_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PlayEffects.DropDown
        cmb_PlayEffects.ItemHeight = 16
    End Sub
    Private Sub cmb_PlayEffects_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PlayEffects.DropDownClosed
        cmb_PlayEffects.ItemHeight = 12
    End Sub


    ' ====================================================================================================
    '  COMBO PLAYER TYPE
    ' ====================================================================================================
    Dim PlayerType As Int32
    Private Sub cmb_PlayerType_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmb_PlayerType.SelectedIndexChanged
        PlayingNotes_ReleaseAllNotes()
        ResetAllSlots()
        SendBendMessage(0.5)
        DynamicVolume = 1
        PlayerType = cmb_PlayerType.SelectedIndex
        Select Case PlayerType
            Case 0
                Label_VolumeSlot.Enabled = True
                txt_VolumeSlot.Enabled = True
                chk_VolumeNegative.Enabled = True
                cmb_PlayEffects.ForeColor = Color.Black
                cmb_PlayEffects.Enabled = True
                Label_Amount.Enabled = True
                txt_Amount.Enabled = True
                SlidersEnable()
            Case 1
                Label_VolumeSlot.Enabled = True
                txt_VolumeSlot.Enabled = True
                chk_VolumeNegative.Enabled = True
                cmb_PlayEffects.ForeColor = Color.Gray
                cmb_PlayEffects.Enabled = False
                Label_Amount.Enabled = False
                txt_Amount.Enabled = False
                SlidersEnable()
            Case 2, 3
                Label_VolumeSlot.Enabled = False
                txt_VolumeSlot.Enabled = False
                chk_VolumeNegative.Enabled = False
                cmb_PlayEffects.ForeColor = Color.Gray
                cmb_PlayEffects.Enabled = False
                Label_Amount.Enabled = False
                txt_Amount.Enabled = False
                SlidersDisable()
            Case 4
                Label_VolumeSlot.Enabled = False
                txt_VolumeSlot.Enabled = False
                chk_VolumeNegative.Enabled = False
                cmb_PlayEffects.ForeColor = Color.Gray
                cmb_PlayEffects.Enabled = False
                Label_Amount.Enabled = True
                txt_Amount.Enabled = True
                SlidersDisable()
        End Select
    End Sub
    Private Sub cmb_PlayerType_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PlayerType.DropDown
        cmb_PlayerType.ItemHeight = 16
    End Sub
    Private Sub cmb_PlayerType_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PlayerType.DropDownClosed
        cmb_PlayerType.ItemHeight = 12
    End Sub


    ' ====================================================================================================
    '  Slider 
    ' ====================================================================================================
    Private Sub SlidersEnable()
        pianoControl1.Top = ClientRectangle.Bottom - 52
        pianoControl1.Height = 50
        pianoControl1.Refresh()
        LabelVol.Show()
        SliderVol.Show()
        LabelPitch.Show()
        SliderPitch.Show()
    End Sub
    Private Sub SlidersDisable()
        LabelVol.Hide()
        SliderVol.Hide()
        LabelPitch.Hide()
        SliderPitch.Hide()
        pianoControl1.Top = ClientRectangle.Bottom - 64
        pianoControl1.Height = 60
        pianoControl1.Refresh()
    End Sub

    Private SliderVolPressed As Boolean
    Private Sub SliderVolFill(ByVal value As Single)
        FillSliderIndicator(SliderVol, value, Color.Orange, Color.AliceBlue)
    End Sub
    Private Sub SliderVol_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles SliderVol.MouseDown
        Cursor.Clip = SliderVol.RectangleToScreen(SliderVol.ClientRectangle)
        SliderVolPressed = True
        SliderVol_Play(e.X)
    End Sub
    Private Sub SliderVol_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles SliderVol.MouseMove
        If Not SliderVolPressed Then Return
        SliderVol_Play(e.X)
    End Sub
    Private Sub SliderVol_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles SliderVol.MouseUp
        Cursor.Clip = Nothing
        SliderVolPressed = False
        'SliderVol_Play(0)
    End Sub
    Private Sub SliderVol_Play(ByVal v As Single)
        v = v / SliderVol.ClientSize.Width
        If VolumeSlot < 0 Then
            Volume = v
        Else
            If chk_VolumeNegative.Checked Then
                Slots.WriteSlot(VolumeSlot, (1 - v) * 1000.0F)
            Else
                Slots.WriteSlot(VolumeSlot, v * 1000.0F)
            End If
        End If
    End Sub

    Private SliderPitchPressed As Boolean
    Private Sub SliderPitchFill(ByVal value As Single)
        FillSliderIndicator(SliderPitch, value, Color.Orange, Color.AliceBlue)
    End Sub
    Private Sub SliderPitch_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles SliderPitch.MouseDown
        Cursor.Clip = SliderPitch.RectangleToScreen(SliderPitch.ClientRectangle)
        SliderPitchPressed = True
        SliderPitch_Play(e.X)
    End Sub
    Private Sub SliderPitch_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles SliderPitch.MouseMove
        If Not SliderPitchPressed Then Return
        SliderPitch_Play(e.X)
    End Sub
    Private Sub SliderPitch_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles SliderPitch.MouseUp
        Cursor.Clip = Nothing
        SliderPitchPressed = False
        SliderPitch_Play(0)
    End Sub
    Private Sub SliderPitch_Play(ByVal v As Single)
        v = v / SliderPitch.ClientSize.Width
        Slots.WriteSlot(NoteSlot, v * 1000.0F)
    End Sub


    ' ====================================================================================================
    '  Playing notes array
    ' ====================================================================================================
    Private PlayingNotes As Hashtable = New Hashtable
    Private Sub PlayingNotes_ReleaseAllNotes()
        For n As Int32 = 0 To 127
            PlayingNotes_Release(n)
        Next
    End Sub
    Private Sub PlayingNotes_Release(ByVal n As Int32)
        If PlayingNotes.Contains(n) Then
            PlayingNotes.Remove(n)
            Dim message As ChannelMessage = New ChannelMessage(ChannelCommand.NoteOff, 0, n)
            synth1.ProcessChannelMessage(message)
            MidiOutSendMessage(message)
            EventsAreEnabled = False
            pianoControl1.ReleasePianoKey(n)
            EventsAreEnabled = True
        End If
    End Sub
    ' -------------------------------------------------------------------- Press - only a initial trigger 
    Private Sub PlayingNotes_Press(ByVal n As Int32, ByVal vol As Int32)
        If Not PlayingNotes.Contains(n) Then
            PlayingNotes.Add(n, 0)
            If vol < 0 Then vol = 0
            If vol > 127 Then vol = 127
            Dim message As ChannelMessage = New ChannelMessage(ChannelCommand.NoteOn, 0, n, vol)
            synth1.ProcessChannelMessage(message)
            MidiOutSendMessage(message)
            EventsAreEnabled = False
            pianoControl1.PressPianoKey(n)
            EventsAreEnabled = True
        End If
    End Sub
    ' ----------------------------------------------------------- Retrigger - changing valume of notes ON
    Private Sub PlayingNotes_Retrigger(ByVal n As Int32, ByVal vol As Int32)
        If vol < 0 Then vol = 0
        If vol > 127 Then vol = 127
        Dim message As ChannelMessage = New ChannelMessage(ChannelCommand.NoteOn, 0, n, vol)
        synth1.ProcessChannelMessage(message)
        MidiOutSendMessage(message)
        '
        If Not PlayingNotes.Contains(n) Then
            PlayingNotes.Add(n, 0)
            EventsAreEnabled = False
            pianoControl1.PressPianoKey(n)
            EventsAreEnabled = True
        End If
    End Sub

    ' ====================================================================================================
    '   OUTPUT MIDI DEVICE
    ' ====================================================================================================
    Friend OutputDeviceID As Int32 = -1
    Private MidiOutDevice As OutputDevice
    Private Function OutputDeviceName(ByVal id As Int32) As String
        If id < 0 Or id >= OutputDevice.DeviceCount Then Return "Unused"
        Dim cap As MidiOutCaps = OutputDevice.GetDeviceCapabilities(id)
        Return cap.name
    End Function
    Private Sub ShowOutDeviceName()
        Combo_Init(cmb_OutputDevice, OutputDeviceName(OutputDeviceID))
    End Sub
    Private Sub OpenOutputDevice()
        CloseOutputDevice()
        If OutputDeviceID < 0 Or OutputDeviceID >= OutputDevice.DeviceCount Then Return
        MidiOutDevice = New OutputDevice(OutputDeviceID)
    End Sub
    Private Sub CloseOutputDevice()
        If MidiOutDevice IsNot Nothing Then
            MidiOutDevice.Close()
            MidiOutDevice = Nothing
        End If
    End Sub
    ' ====================================================================================================
    '   OUTPUT DEVICE COMBO
    ' ====================================================================================================
    Private Sub cmb_OutputDevice_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_OutputDevice.DropDown
        Dim oldname As String = cmb_OutputDevice.Items(0).ToString
        cmb_OutputDevice.ItemHeight = 15
        cmb_OutputDevice.Items.Clear()
        cmb_OutputDevice.Items.Add("Unused")
        For i As Integer = 0 To OutputDevice.DeviceCount - 1
            cmb_OutputDevice.Items.Add(OutputDeviceName(i))
        Next
        If OutputDeviceID >= 0 And OutputDeviceID < cmb_OutputDevice.Items.Count - 1 Then
            cmb_OutputDevice.SelectedIndex = OutputDeviceID + 1
        End If
        Combo_SetIndex_FromString(cmb_OutputDevice, oldname)
    End Sub
    Private Sub cmb_OutputDevice_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_OutputDevice.DropDownClosed
        cmb_OutputDevice.ItemHeight = 10
        Combo_Init(cmb_OutputDevice, cmb_OutputDevice.Text)
    End Sub
    Private Sub cmb_OutputDevice_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_OutputDevice.SelectionChangeCommitted
        OutputDeviceID = cmb_OutputDevice.SelectedIndex - 1
        OpenOutputDevice()
        Save_INI()
    End Sub
    ' ====================================================================================================
    '  MIDI OUT
    ' ====================================================================================================
    Private Sub MidiOutSendMessage(ByVal message As LC.Midi.ChannelMessage)
        If MidiOutDevice Is Nothing Then Return
        Dim m As LC.Midi.ChannelMessage = New LC.Midi.ChannelMessage(message.Command, _
                                                                     OutMidiChannel - 1, _
                                                                     message.Data1, _
                                                                     message.Data2)
        MidiOutDevice.Send(m)
    End Sub



    ' ====================================================================================================
    '  DYNAMIC CONTROLS
    ' ====================================================================================================
    Private kspeed As Single
    Private DynamicVolume As Single = 1
    Private Volume As Single
    Private OldVolume As Single
    Private DeltaVolume As Single
    Private OldNote As Single
    Private DeltaNote As Single
    Private PitchBendNumNotes As Single
    Private CentralNote As Single

    Private VuMeterLeft_Value As Single
    Private VuMeterRight_Value As Single
    Private SliderVolume_Value As Single
    Private SliderPitch_Value As Single
    Private Tuner1_Value As Single
    Private Tuner2_Value As Single

    Private Sub DynamicControls(ByRef note As Single)
        ' ------------------------------------------------------------------ delta volume
        'DeltaVolume = Math.Abs(Volume - OldVolume) / kspeed
        DeltaVolume = (Volume - OldVolume) / kspeed
        OldVolume = Volume
        'Text = DeltaVolume.ToString("0.0")
        ' ------------------------------------------------------------------ delta note
        DeltaNote = pitchBendRangeNnumericUpDown.Value * _
                    Math.Abs(note - OldNote) / kspeed
        OldNote = note
        'Text = DeltaNote.ToString("0.000")
        ' ------------------------------------------------------------------ number of PitchBend notes
        PitchBendNumNotes = 2 * CInt(pitchBendRangeNnumericUpDown.Value)
        ' ------------------------------------------------------------------ snapped is the central note
        If PitchBendNumNotes > 0 Then
            CentralNote = CInt(note * PitchBendNumNotes) / PitchBendNumNotes
        End If
        ' --------------------------------------------------------------------- unassisted tuner value
        Tuner1_Value = 0.5F + (note - CentralNote) * PitchBendNumNotes

        Select Case PlayEffect
            Case 0
                ' ------------------------------------------------------------- no effects
            Case 1, 4
                ' ------------------------------------------------------------- volume reduction
                Dim VolumeMinimum As Single = 1 - (Amount / 100.0F)
                If DeltaNote > 1 Then
                    DynamicVolume -= kspeed * 0.8F
                    If DynamicVolume < VolumeMinimum Then DynamicVolume = VolumeMinimum
                Else
                    DynamicVolume += kspeed * 0.2F
                    If DynamicVolume > 1 Then DynamicVolume = 1
                End If
            Case 2, 5
                ' ------------------------------------------------------------- volume fast change
                DynamicVolume += DeltaVolume * kspeed * 5.0F * (Amount / 100.0F)
                DynamicVolume = 1 + (DynamicVolume - 1) * 0.95F
        End Select
        Select Case PlayEffect
            Case 3, 4, 5
                ' ------------------------------------------------------------- snap amount
                Dim snap As Single
                snap = Math.Abs((note - CentralNote) * 2 * PitchBendNumNotes)
                snap = 1 - CSng(snap ^ (1 + Amount / 25.0F))
                'Text = snap.ToString("0.000")
                ' ------------------------------------------------------------- 0=NoSnap / 1=TotalSnap
                note = CentralNote + (note - CentralNote) * (1 - snap)
                ' ------------------------------------------------------------- Tuner assist
                Tuner2_Value = 0.5F + (note - CentralNote) * PitchBendNumNotes
            Case Else
                Tuner2_Value = 0.5
        End Select
        '
        ' ----------------------------------- test
        'DynamicVolume = 1
        'Debug.Print(DynamicVolume.ToString)
        'Text = DynamicVolume.ToString
        'Text = delta.ToString
    End Sub


    ' ====================================================================================================
    '  POLL SLOTS
    ' ====================================================================================================
    Private Sub PollSlots()
        If Not ToolStripButton_Run.Checked Then Return
        '
        kspeed = PollingTime / 50.0F
        If kspeed > 1 Then kspeed = 1
        '
        ' ==================================================================================
        Dim v As Single

        ' ================================================================================== Volume
        If VolumeSlot >= 0 Then
            Volume = Slots.ReadSlot_NoNan(VolumeSlot) / 1000.0F
            If Volume < 0 Then Volume = 0
            If Volume > 1 Then Volume = 1
            If chk_VolumeNegative.Checked Then Volume = 1 - Volume
        End If

        ' ================================================================================== Players
        If NoteSlot >= 0 Then
            Select Case PlayerType
                ' ========================================================================== THEREMIN CLASSIC
                Case 0
                    v = Slots.ReadSlot_NoNan(NoteSlot) / 1000.0F
                    If v < 0 Then v = 0
                    If v > 1 Then v = 1
                    SliderPitch_Value = v
                    DynamicControls(v)
                    TunerNote = EnabledNotes(0) + CInt((CentralNote - 0.5) * PitchBendNumNotes)
                    SliderVolume_Value = Volume * DynamicVolume
                    ' ----------------------------------------------- Volume to velocity
                    Dim Vnew As Int32 = CInt(Volume * 128 * DynamicVolume)
                    Static Vold As Int32 = -999
                    If Volume > TriggerLevel / 1000.0F And v > 0 Then
                        'If Vnew <> Vold Then
                        '    Vold = Vnew
                        For i As Int32 = 0 To EnabledNotesCount - 1
                            PlayingNotes_Retrigger(EnabledNotes(i), Vnew)
                        Next
                        'End If
                    Else
                    Vold = -999
                    PlayingNotes_ReleaseAllNotes()
                    End If
        ' -----------------------------------------------
        Static oldv As Single = -999
        If Math.Abs(v - oldv) > 0.0001 Then
            oldv = v
            SendBendMessage(v)
        End If
        ' ====================================================================== THEREMIN pizzicato
                Case 1
                    v = Slots.ReadSlot_NoNan(NoteSlot)
                    If v > 1 Then
                        Dim note As Single = v / 1000.0F
                        If note < 0 Then note = 0
                        If note > 1 Then note = 1
                        SliderPitch_Value = note
                        'DynamicControls(note)
                        ThereminPlayer_PlayPizzicato(Volume * DynamicVolume, note)
                    Else
                        ThereminPlayer_Stop()
                    End If
                    SliderVolume_Value = Volume * DynamicVolume
                    Tuner1_Value = 0.5
                    Tuner2_Value = 0.5
                    ' ====================================================================== POLY RETRIGGERED
                Case 2
                    Dim slot As Int32 = NoteSlot
                    For i As Int32 = 0 To EnabledNotesCount - 1
                        v = (Slots.ReadSlot_NoNan(slot)) : slot += 1
                        Dim n As Int32 = EnabledNotes(i)
                        If v > TriggerLevel Then
                            Dim vol As Int32 = CInt(v / 1000.0F * 128)
                            PlayingNotes_Retrigger(n, vol)
                        Else
                            PlayingNotes_Release(n)
                        End If
                    Next
                    Tuner1_Value = 0.5
                    Tuner2_Value = 0.5
                    TunerNote = -1
                    ' ====================================================================== POLYPHONIC CLASSIC
                Case 3
                    Dim slot As Int32 = NoteSlot
                    For i As Int32 = 0 To EnabledNotesCount - 1
                        v = (Slots.ReadSlot_NoNan(slot)) : slot += 1
                        Dim n As Int32 = EnabledNotes(i)
                        If v > TriggerLevel Then
                            Dim vol As Int32 = CInt(v / 1000.0F * 128)
                            PlayingNotes_Press(n, vol)
                        Else
                            PlayingNotes_Release(n)
                        End If
                    Next
                    Tuner1_Value = 0.5
                    Tuner2_Value = 0.5
                    TunerNote = -1
                    ' ====================================================================== RADIOACTIVITY (MCA)
                Case 4
                    Dim slot As Int32 = NoteSlot
                    For i As Int32 = 0 To EnabledNotesCount - 1
                        v = (Slots.ReadSlot_NoNan(slot))
                        Slots.WriteSlot(slot, 0)
                        slot += 1
                        Dim n As Int32 = EnabledNotes(i)
                        If v > TriggerLevel Then
                            Dim vol As Int32 = CInt(v * Amount)
                            PlayingNotes_Release(n)
                            PlayingNotes_Retrigger(n, vol)
                        Else
                            PlayingNotes_Release(n)
                        End If
                    Next
                    Tuner1_Value = 0.5
                    Tuner2_Value = 0.5
                    TunerNote = -1
            End Select
        End If
        '
        ' ================================================================================== VOICE CHANGER PLUS
        Static VoiceSlotPlusTriggered As Boolean = False
        If VoiceSlotPlus >= 0 Then
            v = Slots.ReadSlot_NoNan(VoiceSlotPlus)
            If v < 10000 Then
                If Not VoiceSlotPlusTriggered Then
                    If v > 100 Then
                        SelectNextVoice_ThreadSafe()
                        VoiceSlotPlusTriggered = True
                    End If
                Else
                    If v < 30 Then
                        VoiceSlotPlusTriggered = False
                    End If
                End If
            Else
                v -= 10000
                If v <> VoiceToBeSelected Then
                    VoiceToBeSelected = CInt(v)
                    SelectVoice_ThreadSafe()
                End If
            End If
        End If
        '
        ' ================================================================================== VOICE CHANGER MINUS
        Static VoiceSlotMinusTriggered As Boolean = False
        If VoiceSlotMinus >= 0 Then
            v = Slots.ReadSlot_NoNan(VoiceSlotMinus)
            If Not VoiceSlotMinusTriggered Then
                If v > 100 Then
                    SelectLastVoice_ThreadSafe()
                    VoiceSlotMinusTriggered = True
                End If
            Else
                If v < 30 Then
                    VoiceSlotMinusTriggered = False
                End If
            End If
        End If
        '
        ' ================================================================================== CHORD CHANGER PLUS
        Static ChordSlotPlusTriggered As Boolean = False
        If ChordSlotPlus >= 0 Then
            v = Slots.ReadSlot_NoNan(ChordSlotPlus)
            If v < 10000 Then
                If Not ChordSlotPlusTriggered Then
                    If v > 100 Then
                        NextChord_ThreadSafe()
                        ChordSlotPlusTriggered = True
                    End If
                Else
                    If v < 30 Then
                        ChordSlotPlusTriggered = False
                    End If
                End If
            Else
                v -= 10000
                If v <> ChordToBeSelected Then
                    ChordToBeSelected = CInt(v)
                    SelectChord_ThreadSafe()
                End If
            End If
        End If
        '
        ' ================================================================================== CHORD CHANGER MINUS
        Static ChordSlotMinusTriggered As Boolean = False
        If ChordSlotMinus >= 0 Then
            v = Slots.ReadSlot_NoNan(ChordSlotMinus)
            If Not ChordSlotMinusTriggered Then
                If v > 100 Then
                    LastChord_ThreadSafe()
                    ChordSlotMinusTriggered = True
                End If
            Else
                If v < 30 Then
                    ChordSlotMinusTriggered = False
                End If
            End If
        End If
        '
        ' ================================================================================== CHORD AUTO
        If ChordTime > 0 Then
            Static autochangecounter As Int32
            autochangecounter += CInt(mtimer.Interval)
            If autochangecounter > ChordTime Then
                autochangecounter = 0
                NextChord_ThreadSafe()
            End If
        End If
    End Sub


    ' ====================================================================================================
    '  MILLI TIMER
    ' ====================================================================================================
    Private Sub MilliTimerElapsed_ThreadSafe() Handles mtimer.MilliTimerElapsed
        'Me.Invoke(New MilliTimerElapsed_Delegate(AddressOf MilliTimerElapsed))
        PollSlots()
    End Sub
    'Private Delegate Sub MilliTimerElapsed_Delegate()
    'Private Sub MilliTimerElapsed()
    '    PollSlots()
    'End Sub

    ' ====================================================================================================
    '  USER INTERFACE TIMER
    ' ====================================================================================================
    Private Sub Timer1_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Timer1.Tick
        If Not EventsAreEnabled Then Return
        Vumeter1.Draw(VuMeterLeft_Value)
        Vumeter2.Draw(VuMeterRight_Value)
        SliderVolFill(SliderVolume_Value)
        SliderPitchFill(SliderPitch_Value)
        Tuner1.Draw(Tuner1_Value)
        Tuner2.Draw(Tuner2_Value)
        DrawTunerNote()
    End Sub

    Private TunerNote As Int32
    Private TunerNote_Old As Int32
    Private Sub DrawTunerNote()
        If TunerNote = TunerNote_Old Then Return
        TunerNote_Old = TunerNote
        If TunerNote >= 0 Then
            Label_TunerNote.Text = GetNoteName(TunerNote)
            Label_TunerOctave.Text = GetNoteOctave(TunerNote)
        Else
            Label_TunerNote.Text = ""
            Label_TunerOctave.Text = ""
        End If
    End Sub
    Private Function GetNoteName(ByVal n As Int32) As String
        Select Case n Mod 12
            Case 0 : Return "C"
            Case 1 : Return "C#"
            Case 2 : Return "D"
            Case 3 : Return "D#"
            Case 4 : Return "E"
            Case 5 : Return "F"
            Case 6 : Return "F#"
            Case 7 : Return "G"
            Case 8 : Return "G#"
            Case 9 : Return "A"
            Case 10 : Return "A#"
            Case 11 : Return "B"
            Case Else : Return ""
        End Select
    End Function
    Private Function GetNoteOctave(ByVal n As Int32) As String
        Return (n \ 12).ToString
    End Function

End Class